from pathlib import Path

import solara
from solara.website.components.mailchimp import MailChimp
from solara.website.components.notebook import Notebook

HERE = Path(__file__).parent
title = "Jupyter Dashboard (1/3)"


@solara.component
def Page():
    title = "Build your Jupyter dashboard using Solara"
    solara.Meta(property="og:title", content=title)
    solara.Meta(name="twitter:title", content=title)
    solara.Title(title)

    img = "https://dxhl76zpt6fap.cloudfront.net/public/docs/tutorial/jupyter-dashboard1.webp"
    solara.Meta(name="twitter:image", content=img)
    solara.Meta(property="og:image", content=img)

    description = "Learn how to build a Jupyter dashboard and deploy it as a web app using Solara."
    solara.Meta(name="description", property="og:description", content=description)
    solara.Meta(name="twitter:description", content=description)
    tags = [
        "jupyter",
        "jupyter dashboard",
        "dashboard",
        "web app",
        "deploy",
        "solara",
    ]
    solara.Meta(name="keywords", content=", ".join(tags))

    Notebook(
        Path(HERE / "_jupyter_dashboard_1.ipynb"),
        show_last_expressions=True,
        execute=False,
        outputs={
            "a7d17a84": None,  # empty output (7)
            # original: https://github.com/widgetti/solara/assets/1765949/e844acdb-c77d-4df4-ba4c-a629f92f18a3
            "82f1d2f7": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/map.webp"),  # map (11)
            "3e7ea361": None,  # (13)
            # original: https://github.com/widgetti/solara/assets/1765949/daaa3a46-61f5-431f-8003-b42b5915da4b
            "56055643": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/view.webp"),  # View (15)
            # original: https://github.com/widgetti/solara/assets/1765949/2f4daf0f-b7d8-4f70-b04a-c27542cffdb0
            "c78010ec": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/page.webp"),  # Page (20)
            # original: https://github.com/widgetti/solara/assets/1765949/a691d9f1-f07b-4e06-b21b-20980476ad64
            "18290364": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/controls.webp"),  # Controls
            "0ca68fe8": None,
            "fef5d187": None,
            # original: https://github.com/widgetti/solara/assets/1765949/f0075ad1-808d-458c-8797-e460ce4dc06d
            "af686391": solara.Image("https://dxhl76zpt6fap.cloudfront.net/pages/docs/content/60-jupyter-dashboard-part1/full-app.webp"),  # Full app
        },
    )
    solara.Markdown(
        """
        Explore this app live at [solara.dev](/apps/jupyter-dashboard-1).

        Don’t miss the next tutorial and stay updated with the latest techniques and insights by subscribing to our newsletter.
    """
    )
    location = solara.use_router().path
    MailChimp(location=location)



import os
import shutil
import site
import sys
import threading
import time
import typing
import webbrowser
from enum import Enum
from pathlib import Path

import rich
import rich_click as click
import uvicorn
from rich import print as rprint
from uvicorn.main import LEVEL_CHOICES, LOOP_CHOICES

import solara
from solara.server import settings

from .server import telemetry

try:
    from solara_enterprise.ssg import ssg_crawl
except ImportError:

    def ssg_crawl(*args, **kwargs):  # type: ignore
        raise RuntimeError('SSG not available, please install solara-enterprise (pip install "solara-enterprise[ssg]"')


HERE = Path(__file__).parent

LOGGING_CONFIG: dict = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(levelprefix)s %(message)s",
            "use_colors": None,
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "fmt": '%(levelprefix)s %(client_addr)s - "%(request_line)s" %(status_code)s',  # noqa: E501
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "rich": {
            "class": "rich.logging.RichHandler",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "solara": {"handlers": ["default"], "level": "ERROR"},
        "reacton": {"handlers": ["default"], "level": "ERROR"},
        # "react": {"handlers": ["rich"], "level": "DEBUG"},
        "uvicorn": {"handlers": ["default"], "level": "ERROR"},
        "uvicorn.error": {"level": "ERROR"},
        "uvicorn.access": {"handlers": ["access"], "level": "ERROR", "propagate": False},
    },
}


def _check_version():
    import requests

    try:
        response = requests.get("https://pypi.org/pypi/solara/json")
        latest_version = response.json()["info"]["version"]
    except:  # noqa: E722
        return
    if latest_version != solara.__version__:
        print(f"New version of Solara available: {latest_version}. You have {solara.__version__}. Please upgrade using:")  # noqa: T201
        print(f'\t$ pip install "solara=={latest_version}"')  # noqa: T201


def find_all_packages_paths():
    paths = []
    # sitepackages = set([os.path.dirname(k) for k in site.getsitepackages()])
    sitepackages = {k for k in site.getsitepackages()}
    paths.extend(list(sitepackages))
    for name, module in sys.modules.items():
        if hasattr(module, "__path__"):
            try:
                path = module.__path__[0]
            except:  # noqa: E722
                pass  # happens for namespace packages it seems
                # print(f"Error for {name}")
                # if path:
                #     skip = False
                #     for sitepackage in sitepackages:
                #         if path.startswith(sitepackage):
                #             skip = True
                # if not skip:
                # print(name, path, skip)
                paths.append(str(path))
    # print("PATHS", paths)
    return paths


class EnumType(click.Choice):
    def __init__(self, enum: typing.Type[Enum], case_sensitive=False):
        self._enum = enum
        super().__init__(choices=[item.name for item in enum], case_sensitive=case_sensitive)


@click.group()
def cli():
    pass


production_default = False
if "SOLARA_MODE" in os.environ:
    # settings.main.mode by default is set to production,
    # which is a good default for when you embed in a flask
    # app for instance, but not for the CLI, which app developers
    # usually run.
    production_default = settings.main.mode == "production"
    # Note that in the CLI we do set this value to "development"
    # or "production" based on the --production flag


@cli.command()
@click.option("--port", default=int(os.environ.get("PORT", 8765)))
@click.option(
    "--host",
    default=settings.main.host,
    help="Host to listen on. Defaults to the $HOST environment or $SOLARA_HOST when available or localhost when not given.",
)
@click.option("--dev/--no-dev", default=None, help="Deprecated: use --auto-restart/-a", hidden=True)
@click.option("--production", is_flag=True, default=production_default, help="Run in production mode: https://solara.dev/docs/understanding/solara-server")
@click.option("--reload", is_flag=True, default=None, help="Deprecated: use --auto-restart/-a", hidden=True)
@click.option("-a", "--auto-restart", is_flag=True, default=False, help="Enable auto-restarting of server when the solara server code changes.")
@click.option("--tracer/--no-tracer", default=False)
@click.option("--timing/--no-timing", default=False)
@click.option("--open/--no-open", default=True)
@click.option(
    "--restart-dir",
    "restart_dirs",
    multiple=True,
    help="Set restart directories explicitly, instead of using the current working" " directory.",
    type=click.Path(exists=True),
)
@click.option(
    "--restart-exclude",
    "restart_excludes",
    multiple=True,
    help="Set glob patterns to exclude while watching for files. Includes "
    "'.*, .py[cod], .sw.*, ~*' by default; these defaults can be overridden "
    "with `--restart-include`. This option has no effect unless watchgod is "
    "installed.",
)
@click.option(
    "--workers",
    default=None,
    type=int,
    help="Number of worker processes. Defaults to the $WEB_CONCURRENCY environment" " variable if available, or 1. Not valid with --auto-restart/-a.",
)
@click.option(
    "--env-file",
    type=click.Path(exists=True),
    default=None,
    help="Environment configuration file.",
    show_default=True,
)
@click.option(
    "--log-config",
    type=click.Path(exists=True),
    default=None,
    help="Logging configuration file. Supported formats: .ini, .json, .yaml.",
    show_default=True,
)
@click.option(
    "--log-level",
    type=LEVEL_CHOICES,
    default=None,
    help="Log level. [default: info]",
    show_default=True,
)
@click.option(
    "--log-level-uvicorn",
    type=LEVEL_CHOICES,
    default="error",
    help="Log level. [default: error]",
    show_default=True,
)
@click.option(
    "--access-log/--no-access-log",
    is_flag=True,
    default=True,
    help="Enable/Disable access log.",
)
@click.option(
    "--root-path",
    type=str,
    default="",
    help="Set the ASGI 'root_path' for applications submounted below a given URL path.",
)
@click.option(
    "--theme-loader",
    type=str,
    default=settings.theme.loader,
    help=f"Loader to use when the app is not yet shown to the user. [default: {settings.theme.loader!r}]",
)
@click.option(
    "--theme-variant",
    type=settings.ThemeVariant,
    default=settings.theme.variant.name,
    help=f"Use light or dark variant, or auto detect (auto). [default: {settings.theme.variant.name}",
)
@click.option(
    "--dark",
    type=bool,
    default=settings.theme.variant == settings.ThemeVariant.dark,
    help="Use dark theme. Shorthand for --theme-variant=dark",
)
@click.option(
    "--theme-variant-user-selectable/--no-theme-variant-user-selectable",
    type=bool,
    hidden=True,
    help="Deprecated.",
)
@click.option("--pdb/--no-pdb", "use_pdb", default=False, help="Enter debugger on error")
@click.argument("app")
@click.option(
    "--loop",
    type=LOOP_CHOICES,
    default="auto",
    help="Event loop implementation.",
    show_default=True,
)
@click.option(
    "--ssg/--no-ssg",
    is_flag=True,
    default=settings.ssg.enabled,
    help="(pre) Render static pages.",
)
@click.option(
    "--search/--no-search",
    is_flag=True,
    default=settings.search.enabled,
    help="Enable search (requires ssg generated pages).",
)
@click.option(
    "--check-version/--no-check-version",
    is_flag=True,
    default=True,
    help="Check installed version again pypi version.",
)
def run(
    app,
    host,
    port,
    open,
    auto_restart: bool,
    reload: bool,
    restart_dirs: typing.Optional[typing.List[str]],
    restart_excludes: typing.List[str],
    dev: bool,
    production: bool,
    tracer: bool,
    timing: bool,
    loop: str,
    workers: int,
    env_file: str,
    root_path: str,
    log_config: str,
    log_level: str,
    log_level_uvicorn: str,
    access_log: bool,
    use_pdb: bool,
    theme_loader: str,
    theme_variant: settings.ThemeVariant,
    dark: bool,
    theme_variant_user_selectable: bool,
    ssg: bool,
    search: bool,
    check_version: bool = True,
):
    """Run a Solara app."""
    if dev is not None:
        print("solara: --dev is deprecated, use --auto-restart/-a instead", file=sys.stderr)  # noqa: T201
        auto_restart = dev
    if reload is not None:
        print("solara: --reload is deprecated, use --auto-restart/-a instead", file=sys.stderr)  # noqa: T201
        auto_restart = reload
    if check_version:
        _check_version()

    # uvicorn calls it reload, we call it auto restart
    reload = auto_restart
    del auto_restart
    settings.ssg.enabled = ssg
    settings.search.enabled = search
    reload_dirs = restart_dirs if restart_dirs else None
    del restart_dirs
    url = f"http://{host}:{port}"

    failed = False
    if reload:
        telemetry._auto_restart_enabled = True
        solara_root = Path(solara.__file__).parent

        reload_dirs = list(reload_dirs if reload_dirs else [])

        # we restart the server when solara or react changes, in principle we should do
        # that for all dependencies of the server, but these are changing most often
        # during development
        # We exclude the website, that will be handled by solara/server/reload.py
        reload_dirs = [str(solara_root), str(Path(solara.__file__).parent)]
        try:
            import solara_enterprise

            reload_dirs.append(str(Path(solara_enterprise.__file__).parent))
            del solara_enterprise
        except ImportError:
            pass
        reload_excludes = restart_excludes if restart_excludes else []
        del restart_excludes
        reload_excludes = [str(solara_root / "website"), str(solara_root / "template")]
        reload_excludes.append(app)
        del solara_root
        reload = True
        # avoid sending many restarts
        settings.telemetry.mixpanel_enable = False
    else:
        del restart_excludes

    if production:
        settings.main.mode = "production"
    else:
        settings.main.mode = "development"
    server = None

    # TODO: we might want to support this, but it needs to be started from the main thread
    # and then uvicorn needs to be started from a thread
    # def open_webview():
    #     import webview

    #     while not failed and (server is None or not server.started):
    #         time.sleep(0.1)
    #     if not failed:
    #         window = webview.create_window("Hello world", url, resizable=True)
    #         window.on_top = True
    #         # window.show()
    #         webview.start(debug=True)

    def open_browser():
        while not failed and (server is None or not server.started):
            time.sleep(0.1)
        if not failed:
            webbrowser.open(url)

    if open:
        threading.Thread(target=open_browser, daemon=True).start()

    rich.print(f"Solara server is starting at {url}")

    if log_level is not None:
        LOGGING_CONFIG["loggers"]["solara"]["level"] = log_level.upper()
        # LOGGING_CONFIG["loggers"]["reacton"]["level"] = log_level.upper()

    log_level = log_level_uvicorn
    del log_level_uvicorn

    kwargs = locals().copy()
    # cgi vars: https://datatracker.ietf.org/doc/html/rfc3875
    os.environ["SOLARA_APP"] = app
    os.environ["SERVER_NAME"] = host
    os.environ["SERVER_PORT"] = str(port)

    kwargs["app"] = "solara.server.starlette:app"
    kwargs["log_config"] = LOGGING_CONFIG if log_config is None else log_config
    kwargs["loop"] = loop
    settings.main.use_pdb = use_pdb
    settings.theme.loader = theme_loader
    if dark:
        theme_variant = settings.ThemeVariant.dark
    settings.theme.variant = theme_variant
    settings.main.tracer = tracer
    settings.main.timing = timing
    items = (
        "theme_variant_user_selectable dark theme_variant theme_loader use_pdb server open_browser open url failed dev tracer"
        " timing ssg search check_version production".split()
    )
    for item in items:
        del kwargs[item]

    def start_server():
        nonlocal server
        nonlocal failed
        try:
            # we manually create the server instead of calling uvicorn.run
            # because we can then access the server variable and check if it is
            # running.
            config = uvicorn.Config(**kwargs)
            server = uvicorn.Server(config=config)
            if reload:
                sock = config.bind_socket()
                from uvicorn.supervisors import ChangeReload

                ChangeReload(
                    config,
                    target=run_with_settings(
                        server, main=settings.main.dict(), theme=settings.theme.dict(), ssg=settings.ssg.dict(), search=settings.search.dict()
                    ),
                    sockets=[sock],
                ).run()
            else:
                server.run()
        except:  # noqa
            failed = True
            raise

    def ssg_run():
        while not failed and (server is None or not server.started):
            time.sleep(0.1)
        if not failed:
            assert server is not None
            base_url = f"http://{server.config.host}:{server.config.port}"
            rprint("Running Static Site Generator pre-render background task")
            ssg_crawl(base_url)
            if settings.search.enabled:
                from solara_enterprise.search.index import build_index

                build_index("")

    # in dev mode we run the ssg in the child process (see run_with_settings)
    if not dev and settings.ssg.enabled:
        threading.Thread(target=ssg_run, daemon=True).start()

    # if we don't have to wait for SSG, we can build the index right away
    if not settings.ssg.enabled and settings.search.enabled:
        from solara_enterprise.search.index import build_index

        build_index("")

    start_server()

    # TODO: if we want to use webview, it should be sth like this
    # server_thread = threading.Thread(target=start_server)
    # server_thread.start()
    # if open:
    #     # open_webview()
    #     open_browser()
    # server_thread.join()


@cli.command()
@click.argument("app")
@click.option("--port", default=int(os.environ.get("PORT", 8765)))
@click.option("--host", default=settings.main.host)
@click.option(
    "--headed/--no-headed",
    is_flag=True,
    default=settings.ssg.headed,
    help="Show browser window if true.",
)
def ssg(app: str, port: int, host: str, headed: bool):
    """Static site generation"""
    settings.ssg.headed = headed
    settings.ssg.enabled = True
    settings.main.mode = "production"  # always override this
    os.environ["SOLARA_APP"] = app
    from solara.server.starlette import ServerStarlette

    server = ServerStarlette(port=port, host=host)
    server.serve_threaded()
    server.wait_until_serving()

    base_url = f"http://{server.server.config.host}:{server.server.config.port}"

    ssg_crawl(base_url)


@cli.command()
def deploy():
    rprint("...")
    import time

    time.sleep(1)
    rprint("Want your app to run instantly on awesomeapp-mystartup-gh.solara.run?")
    rprint("\tCheck out https://solara.dev/docs/deploying/cloud-hosted")


@cli.command()
@click.argument("app")
def search(app: str):
    """Build search index based on ssg output"""
    os.environ["SOLARA_APP"] = app
    from solara_enterprise.search.index import build_index

    build_index("")


class run_with_settings:
    """This cross a process boundary, and takes the serialized settings with it"""

    def __init__(self, server: uvicorn.Server, main: typing.Dict, theme: typing.Dict, ssg: typing.Dict, search: typing.Dict):
        self.server = server
        self.main = main
        self.theme = theme
        self.ssg = ssg
        self.search = search

    def __call__(self, *args, **kwargs):
        # this is now in the new process, where we need to re-apply the settings
        failed = False

        def ssg_run():
            while not failed and (self.server is None or not self.server.started):
                time.sleep(0.1)
            if not failed:
                assert self.server is not None
                base_url = f"http://{self.server.config.host}:{self.server.config.port}"
                rprint("Running Static Site Generator pre-render background task")
                ssg_crawl(base_url)
                if settings.search.enabled:
                    from solara_enterprise.search.index import build_index

                    build_index("")

        settings.main = settings.MainSettings(**self.main)
        settings.theme = settings.ThemeSettings(**self.theme)
        settings.ssg = settings.SSG(**self.ssg)
        settings.search = settings.Search(**self.search)
        if settings.ssg.enabled:
            threading.Thread(target=ssg_run, daemon=True).start()
        try:
            return self.server.run(*args, **kwargs)
        except:  # noqa
            failed = True
            raise


@cli.command()
@click.option("--port", default=int(os.environ.get("PORT", 8000)))
def staticserve(port):
    """Experimental static serving"""
    import http.server
    import os
    from functools import partial

    print(f"http://localhost:{port}/")  # noqa
    wk_dir = os.getcwd() + "/staticbuild"
    http.server.SimpleHTTPRequestHandler.extensions_map[".js"] = "application/javascript"
    http.server.SimpleHTTPRequestHandler.extensions_map = {k: v + ";charset=UTF-8" for k, v in http.server.SimpleHTTPRequestHandler.extensions_map.items()}
    http.server.test(HandlerClass=partial(http.server.SimpleHTTPRequestHandler, directory=wk_dir), port=port, bind="")  # type: ignore


@cli.command()
def staticbuild():
    """Experimental static build"""
    # imports locals, otherwise .run() does not use the cli arguments
    import solara.server
    import solara.server.server

    server_path = HERE / "server"
    assets_path = server_path / "assets"
    static_path = server_path / "static"

    target_dir = Path("staticbuild")
    target_dir.mkdir(exist_ok=True)

    from .server import settings

    copytree(settings.assets.proxy_cache_dir, target_dir / "_solara/cdn/")

    static_dir_target = target_dir / "static"
    static_dir_target.mkdir(exist_ok=True)

    assets_dir_target = target_dir / "static" / "assets"
    assets_dir_target.mkdir(exist_ok=True)
    copytree(assets_path, assets_dir_target)

    for path in (
        list(static_path.glob("*.css"))
        + list(static_path.glob("*.js"))
        + list(static_path.glob("*.ps"))
        + list(static_path.glob("*.py"))
        + list(static_path.glob("*.png"))
        + list(static_path.glob("*.svg"))
    ):
        shutil.copy(path, static_dir_target)

    include_nbextensions = True
    if include_nbextensions:
        directories = solara.server.server.get_nbextensions_directories()
        nbextensions, ignore = solara.server.server.get_nbextensions()
        for name in nbextensions:
            for directory in directories:
                if (directory / (name + ".js")).exists():
                    src = (directory / (name + ".js")).parent
                    dst: Path = (static_dir_target / "nbextensions" / (name + ".js")).parent
                    dst.mkdir(parents=True, exist_ok=True)
                    # shutil.copytree(src, dst)
                    copytree(src, dst)

    target_dir_wheels = target_dir / "wheels"
    target_dir_wheels.mkdir(exist_ok=True)
    version = solara.__version__
    for path in [Path(f"dist/solara-{version}-py2.py3-none-any.whl")]:
        shutil.copy(path, target_dir_wheels)

    target_dir_static_dist = target_dir / "static/dist"
    target_dir_static_dist.mkdir(exist_ok=True)
    voila = server_path / "dist"
    for path in list(voila.glob("*.js")) + list(voila.glob("*.woff")):
        shutil.copy(path, target_dir_static_dist)

    index_html = solara.server.server.read_root("", render_kwargs={"for_pyodide": True}, use_nbextensions=True)
    assert index_html is not None
    (target_dir / "index.html").write_text(index_html)


@click.group()
def create():
    """Quickly create a solara script or project."""
    pass


@create.command()
@click.argument(
    "target",
    type=click.Path(exists=False),
    default="sol.py",
    required=False,
)
def button(target: typing.Optional[Path]):
    """Create a button with a click counter."""
    write_script("button", target)


@create.command()
@click.argument(
    "target",
    type=click.Path(exists=False),
    default="sol.py",
    required=False,
)
def markdown(target: typing.Optional[Path] = None):
    """Create a markdown editor."""
    write_script("markdown", target)


def write_script(name: str, target: typing.Optional[Path]):
    code = (HERE / "template" / f"{name}.py").read_text()
    if target is None:
        target = Path("sol.py")
    else:
        target = Path(target)
    target.parent.mkdir(exist_ok=True)
    target.write_text(code)
    rprint(f"Wrote:  {target.resolve()}")
    rprint(f"Run as:\n\t $ solara run {target.resolve()}")


# recursively copy a directory and allow for existing directories
def copytree(src: Path, dst: Path, copy_function=shutil.copy2, ignore: typing.Callable[[Path], bool] = lambda x: False, rename=lambda x: x):
    print(src, " -> ", dst)  # noqa
    if not src.exists():
        return
    if not dst.exists():
        dst.mkdir(parents=True, exist_ok=True)
    for item in src.iterdir():
        if ignore(item):
            continue
        if item.is_dir():
            copytree(item, dst / rename(item).name, ignore=ignore, copy_function=copy_function, rename=rename)
        else:
            copy_function(item, dst / rename(item).name)


@create.command()
@click.argument(
    "target",
    type=click.Path(exists=False),
    default="solara_portal",
    required=False,
)
def portal(target: Path):
    """Create a full Python project template for a data portal"""
    target = Path(target)
    name = target.name
    package_name = name.replace("-", "_")

    def copy_function(src: Path, dst: Path):
        dst.write_bytes(
            src.read_bytes().replace(b"solara-portal", name.encode("utf8")).replace(b"solara_portal", package_name.encode("utf8")),
        )

    def rename(path: Path):
        if path.name == "solara_portal":
            return path.parent / package_name
        else:
            return path

    copytree(
        HERE / "template" / "portal",
        target,
        ignore=lambda p: p.name.startswith("__") and p.name != "__init__.py",
        copy_function=copy_function,
        rename=rename,
    )
    rprint(f"Wrote:  {target.resolve()}")
    rprint(f"Install as:\n\t $ (cd {target}; pip install -e .)")
    rprint(f"Run as:\n\t $ solara run {package_name}.pages")


cli.add_command(run)
cli.add_command(staticbuild)
cli.add_command(create)
cli.add_command(ssg)


def main():
    cli()


if __name__ == "__main__":
    main()



from typing import Dict, List, Union

import reacton.ipyvuetify as v

import solara
import solara.util


@solara.component
def ProgressLinear(
    value: Union[bool, float] = True,
    color="primary",
    style: Union[str, Dict[str, str], None] = None,
    classes: List[str] = [],
):
    """Progress bar component showing a percentage, indeterminate or hidden.

     * When `value` is `True`, the progress bar will be indeterminate.
     * When `value` is `False`, the progress bar will be hidden, but still take up space.
       This can be used to avoid the page to jump when the progress bar is shown and hidden.
     * When `value` is a number between 0 and 100, the progress bar will show the percentage.

    ### Basic example

    ```solara
    import solara

    @solara.component
    def Page():
        solara.ProgressLinear(value=30, color="purple")
        solara.ProgressLinear(True)
    ```

    ## Arguments
     * `value`: Value of the progress bar. Can be a number between 0 and 100, `True` for indeterminate or `False` to hide.
     * `color`: Color of the progress bar.
     * `style`: CSS style to add to the progress bar.
     * `classes`: List of classes to apply to the progress bar.

    """
    indeterminate = value is True

    style_flat = solara.util._flatten_style(style)
    if value is False:
        style_flat = "visibility: hidden;" + style_flat
    class_ = solara.util._combine_classes(classes)
    v.ProgressLinear(indeterminate=indeterminate, value=value if value is not True else None, color=color, style_=style_flat, class_=class_)



import solara
from solara.alias import rv

markdown_initial = """
# Large
## Smaller

## List items

    * item 1
    * item 2


## Code highlight support
```python
code = "formatted" and "supports highlighting"
```

    """.strip()


@solara.component
def MarkdownEditor():
    markdown_text, set_markdown_text = solara.use_state(markdown_initial)
    with solara.ColumnsResponsive(12, medium=6):
        with solara.Column():
            solara.Markdown("# Input text")
            with solara.Padding(2):
                with rv.Sheet(elevation=2):
                    rv.Textarea(v_model=markdown_text, on_v_model=set_markdown_text, rows=20)
        with solara.Column():
            solara.Markdown("# Renders like")
            with solara.Padding(2):
                solara.Markdown(markdown_text)
        with solara.Sidebar():
            with solara.Row():
                solara.Button("Clear", on_click=lambda: set_markdown_text(""))
                solara.Button("Reset ", on_click=lambda: set_markdown_text(markdown_initial))


# create an alias of the Markdown Editor component so Solara can find it
Page = MarkdownEditor



import logging
import sys
from pathlib import Path

import ipyvuetify as v
import ipywidgets
import reacton.core

# import pytest
import solara

# import solara.server.app
from solara.server import reload
from solara.server.app import AppScript

logger = logging.getLogger("solara.server.app_test")


HERE = Path(__file__).parent
reload.reloader.start()


def test_notebook_element(kernel_context, no_kernel_context):
    name = str(HERE / "solara_test_apps" / "notebookapp_element.ipynb")
    app = AppScript(name)
    try:
        with kernel_context:
            el = app.run()
            assert isinstance(el, reacton.core.Element)
            el2 = app.run()
            assert el is el2
    finally:
        app.close()


def test_notebook_component(kernel_context, no_kernel_context):
    name = str(HERE / "solara_test_apps" / "notebookapp_component.ipynb")
    app = AppScript(name)
    try:
        with kernel_context:
            el = app.run()
            assert isinstance(el, reacton.core.Element)
            el2 = app.run()
            assert el is el2
    finally:
        app.close()


def test_notebook_widget(kernel_context, no_kernel_context):
    name = str(HERE / "solara_test_apps" / "notebookapp_widget.ipynb")
    app = AppScript(name)
    try:
        with kernel_context:
            el = app.run()
            root = solara.RoutingProvider(children=[el], routes=app.routes, pathname="/")
            _box, rc = solara.render(root, handle_error=False)
            widget1 = rc.find(ipywidgets.Button, description="Click me").widget
            assert isinstance(widget1, ipywidgets.Button)
            _box, rc = solara.render(root, handle_error=False)
            widget2 = rc.find(ipywidgets.Button, description="Click me").widget
        assert widget1 is not widget2
    finally:
        app.close()


def test_sidebar_single_file_multiple_routes(kernel_context, no_kernel_context):
    name = str(HERE / "solara_test_apps" / "single_file_multiple_routes.py")
    app = AppScript(name)
    try:
        with kernel_context:
            c = app.run()
            root = solara.RoutingProvider(children=[c], routes=app.routes, pathname="/")
            box, rc = solara.render(root, handle_error=False)
            assert rc.find(v.Slider, label="in sidebar")
    finally:
        app.close()


def test_sidebar_single_file(kernel_context, no_kernel_context):
    name = str(HERE / "solara_test_apps" / "single_file.py")
    app = AppScript(name)
    try:
        with kernel_context:
            c = app.run()
            root = solara.RoutingProvider(children=[c], routes=app.routes, pathname="/")
            box, rc = solara.render(root, handle_error=False)
            assert rc.find(v.Slider, label="in sidebar")
    finally:
        app.close()


def test_sidebar_single_file_missing(kernel_context, no_kernel_context):
    name = str(HERE / "solara_test_apps" / "single_file.py:doesnotexist")
    app = AppScript(name)
    try:
        with kernel_context:
            c = app.run()
            root = solara.RoutingProvider(children=[c], routes=app.routes, pathname="/")
            box, rc = solara.render(root, handle_error=False)
            assert "No object with name doesnotexist" in rc.find(v.Alert).widget.children[0]
    finally:
        app.close()


# these make other test fail on CI (vaex is used, which causes a blake3 reload, which fails)
def test_watch_module_reload(tmpdir, kernel_context, extra_include_path, no_kernel_context):
    import ipyvuetify as v

    with extra_include_path(str(tmpdir)):
        py_file = tmpdir / "test.py"
        py_mod_file = tmpdir / "somemod.py"

        logger.info("writing files")
        with open(py_mod_file, "w") as f:
            f.write("import ipyvuetify as v; page = v.Btn.element(children=['first'])\n")
        with open(py_file, "w") as f:
            f.write("import somemod; page=somemod.page\n")

        logger.info("wrote files")

        app = AppScript(f"{py_file}")
        try:
            result = app.run()
            assert "somemod" in sys.modules
            assert "somemod" in reload.reloader.watched_modules
            somemod1 = sys.modules["somemod"]
            root = solara.RoutingProvider(children=[result], routes=app.routes, pathname="/")
            box, rc = solara.render(root, handle_error=False)
            assert rc.find(v.Btn, children=["first"])
            # assert result.component.widget == v.Btn

            # change depending module
            with open(py_mod_file, "w") as f:
                f.write("import ipyvuetify as v; page = v.Card.element(children=['second'])\n")
            # wait for the event to trigger
            reload.reloader.reload_event_next.wait()
            # assert "somemod" not in sys.modules
            # breakpoint()
            result = app.run()
            assert "somemod" in sys.modules
            root = solara.RoutingProvider(children=[result], routes=app.routes, pathname="/")
            box, rc = solara.render(root, handle_error=False)
            assert rc.find(v.Card, children=["second"])
            somemod2 = sys.modules["somemod"]
            assert somemod1 is not somemod2
        finally:
            app.close()
            if "somemod" in sys.modules:
                del sys.modules["somemod"]
            reload.reloader.watched_modules.remove("somemod")


# def test_script_reload_component(tmpdir, kernel_context, extra_include_path, no_kernel_context):
#     import ipyvuetify as v

#     with extra_include_path(str(tmpdir)):
#         py_file = tmpdir / "test.py"

#         logger.info("writing files")
#         with open(py_file, "w") as f:
#             f.write("import reacton.ipyvuetify as v; Page = v.Btn\n")

#         app = AppScript(f"{py_file}")
#         try:
#             result = app.run()
#             assert result().component.widget == v.Btn
#             with open(py_file, "w") as f:
#                 f.write("import reacton.ipyvuetify as v; Page = v.Slider\n")
#             # wait for the event to trigger
#             reload.reloader.reload_event_next.wait()
#             # assert "somemod" not in sys.modules
#             # breakpoint()
#             result = app.run()
#             assert result().component.widget == v.Slider
#         finally:
#             app.close()


# def test_watch_module_import_error(tmpdir, kernel_context, extra_include_path, no_kernel_context):
#     import ipyvuetify as v

#     with extra_include_path(str(tmpdir)):
#         py_file = tmpdir / "test.py"
#         py_mod_file = tmpdir / "somemod2.py"

#         logger.info("writing files")
#         with open(py_mod_file, "w") as f:
#             f.write("import ipyvuetify as v; App = v.Btn.element\n")
#         with open(py_file, "w") as f:
#             f.write("import somemod2; app=somemod2.App\n")

#         logger.info("wrote files")

#         app = AppScript(f"{py_file}")
#         try:
#             result = app.run()
#             assert "somemod2" in sys.modules
#             assert "somemod2" in reload.reloader.watched_modules
#             assert result().component.widget == v.Btn

#             # syntax error
#             with open(py_mod_file, "w") as f:
#                 f.write("import ipyvuetify as v; App !%#$@= v.Card.element\n")
#             reload.reloader.reload_event_next.wait()
#             with pytest.raises(SyntaxError):
#                 result = app.run()

#             with open(py_mod_file, "w") as f:
#                 f.write("import ipyvuetify as v; App = v.Card.element\n")
#             reload.reloader.reload_event_next.wait()
#             result = app.run()
#             assert "somemod2" in sys.modules
#             assert result().component.widget == v.Card
#         finally:
#             app.close()
#             del sys.modules["somemod2"]
#             reload.reloader.watched_modules.remove("somemod2")



from typing import Callable, Dict, List, Optional, TypeVar, Union

import solara
from solara import v


@solara.component
def Tab(
    label: Union[None, str, solara.Element] = None,
    icon_name: Optional[str] = None,
    path_or_route: Union[None, str, "solara.Route"] = None,
    disabled=False,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
    children: List[solara.Element] = [],
    tab_children: List[Union[solara.Element, str]] = [],
):
    """An item in a Tabs component.

    (*Note: [This component is experimental and its API may change in the future](/docs/lab).*)

    Should be a direct child of a [Tabs](/api/tabs).

    ## Arguments
     * `label`: The label of the tab.
     * `icon_name`: The name of the icon to display in the tab.
     * `path_or_route`: The path or route to navigate to when the tab is clicked.
     * `disabled`: Whether the tab is disabled.
     * `classes`: Additional CSS classes to apply.
     * `style`: CSS style to apply.
     * `children`: The children of the tab. These will be displayed when the tab is active.
     * `tab_children`: The children of the tab header. These will be displayed in the tab
        header, if a label or icon_name is provided they are prepended to the `tab_children`.

    """
    if label is not None:
        tab_children = [label] + tab_children
    if icon_name:
        tab_children = [v.Icon(left=bool(label), children=[icon_name])] + tab_children
    style_flat = solara.util._flatten_style(style)
    class_ = solara.util._combine_classes(classes)
    # note: children is not used, it is only used in the Tabs component
    return v.Tab(children=tab_children, disabled=disabled, class_=class_, style_=style_flat)


T = TypeVar("T")


@solara.component
def Tabs(
    value: Union[None, int, "solara.Reactive[int]"] = None,
    on_value: Optional[Callable[[int], None]] = None,
    color: Optional[str] = None,
    background_color: Optional[str] = None,
    slider_color: Optional[str] = None,
    dark: bool = False,
    grow: bool = False,
    vertical=False,
    align: str = "left",
    lazy=False,
    children: List[solara.Element] = [],
):
    """A tabbed container showing one tab at a time.

    (*Note: [This component is experimental and its API may change in the future](/docs/lab).*)

    Note that if Tabs are used as a child of the [AppBar](/api/appbar) component, the tabs
    will be placed under the app bar. See our [authorization app](/apps/authorization) for an example.

    If the children [Tab](/api/tab) elements are passed a `path_or_route` argument, the active tab
    will be based on the path of the current page.


    ## Examples

    ### Only tabs headers

    ```solara
    import solara
    import solara.lab


    @solara.component
    def Page():
        with solara.lab.Tabs():
            solara.lab.Tab("Tab 1")
            solara.lab.Tab("Tab 2")
    ```

    ### Tabs with content

    This is usually only used when the tabs are placed in the [AppBar](/api/appbar) component.

    ```solara
    import solara
    import solara.lab


    @solara.component
    def Page():
        with solara.lab.Tabs():
            with solara.lab.Tab("Tab 1"):
                solara.Markdown("Hello")
            with solara.lab.Tab("Tab 2"):
                solara.Markdown("World")
    ```


    ### Tabs events

    The `value` on the Tabs component is a reactive value that can be used to
    listen to changes in the selected tab and make the UI respond to it.

    ```solara
    import solara
    import solara.lab

    tab_index = solara.reactive(0)


    @solara.component
    def Page():

        def next_tab():
            tab_index.value = (tab_index.value + 1) % 2

        solara.Title(f"Tab {tab_index.value + 1}")
        solara.Button('Next Tab', on_click=next_tab)

        with solara.lab.Tabs(value=tab_index):
            with solara.lab.Tab("Tab 1"):
                solara.Markdown("Hello")
            with solara.lab.Tab("Tab 2"):
                solara.Markdown("World")
            with solara.lab.Tab("Disabled", disabled=True):
                solara.Markdown("World")

    ```

    ### Advanced tabs

    Tabs can be nested, styled and placed vertically.

    ```solara
    import solara
    import solara.lab


    @solara.component
    def Page():
        with solara.lab.Tabs(background_color="primary", dark=True):
            with solara.lab.Tab("Home", icon_name="mdi-home"):
                solara.Markdown("Hello")
            with solara.lab.Tab("Advanced", icon_name="mdi-apps"):
                with solara.lab.Tabs(grow=True, background_color="primary", dark=True, slider_color="green"):
                    with solara.lab.Tab("Settings", icon_name="mdi-cogs"):
                        with solara.lab.Tabs(vertical=True, slider_color="green"):
                            with solara.lab.Tab("User", icon_name="mdi-account"):
                                solara.Markdown("User settings")
                            with solara.lab.Tab("System", icon_name="mdi-access-point"):
                                solara.Markdown("System settings")
                    with solara.lab.Tab("Analytics", icon_name="mdi-chart-line"):
                        with solara.lab.Tabs(vertical=True):
                            with solara.lab.Tab("User", icon_name="mdi-account"):
                                solara.Markdown("User analytics")
                            with solara.lab.Tab("System", icon_name="mdi-access-point"):
                                solara.Markdown("System analytics")

    ```


    ### Many tabs

    If many tabs are shown, paginations arrows are shown.

    ```solara
    import solara
    import solara.lab

    tab_count = 30


    @solara.component
    def Page():
        with solara.lab.Tabs():
            for i in range(tab_count):
                with solara.lab.Tab(f"Tab {i+1}"):
                    solara.Markdown(f"Content for tab {i+1}")
    ```


    ## Arguments

     * `value`: The index of the selected tab. If `None`, the first tab is selected or it is based in the route/path.
     * `on_value`: A callback that is called when the selected tab changes.
     * `color`: The color of text in the tab headers (only for dark=False).
     * `background_color`: The background color of the tab headers.
     * `slider_color`: The color of the slider.
     * `dark`: Apply a dark theme.
     * `grow`: Whether the tabs should grow to fill the available space.
     * `vertical`: Whether the tabs are vertical.
     * `align`: The alignment of the tabs, possible values are 'left', 'start', 'center', 'right' or 'end'.
     * `lazy`: Whether the child components of the inactive tabs are rendered or not. If lazy=True, components of inactive tabs are not rendered.
     * `classes`: Additional CSS classes to apply.
     * `style`: CSS style to apply.
    """

    paths_of_routes = [child.kwargs.get("path_or_route") for child in children]
    paths = [solara.resolve_path(path_or_route, level=0) if path_or_route else None for path_or_route in paths_of_routes]
    router = solara.use_router()
    if value is None:
        if router.path in paths:
            value = paths.index(router.path)
        else:
            value = 0

    def safe_on_value(index: Optional[int]):
        if on_value and index is not None:
            on_value(index)

    reactive_value = solara.use_reactive(value, safe_on_value)
    del value

    has_content = False
    for i, child in enumerate(children):
        if not child.component == Tab:
            raise ValueError(f"Tabs children must be Tab components, but child {i} is {child.component}")
        if child.kwargs.get("children"):
            has_content = True

    def on_v_model(index: Optional[int]):
        if index is not None:
            path = paths[index]
            if path:
                router.push(path)
            reactive_value.value = index

    if align not in ["left", "start", "center", "right", "end"]:
        raise ValueError(f"Tabs align must be one of 'left', 'start', 'center', 'right', 'end', but is {align}")

    with v.Tabs(
        v_model=reactive_value.value,
        on_v_model=on_v_model,
        centered=align == "center",
        right=align in ["right", "end"],
        children=children,
        vertical=vertical,
        color=color,
        background_color=background_color,
        show_arrows=True,
        grow=grow,
        dark=dark,
    ) as tabs:
        v.TabsSlider(color=slider_color)
        if has_content:
            with v.TabsItems(v_model=reactive_value.value, on_v_model=on_v_model):
                for i, child in enumerate(children):
                    if not lazy or reactive_value.value == i:
                        v.TabItem(children=child.kwargs.get("children", []), value=i)
                    else:
                        v.TabItem(
                            value=i,
                            children=[
                                # Nice idea, but by using the widget interface the tab does not change without binding using
                                # v-model. So we would need to implement this using a vuetify template.
                                # v.SkeletonLoader(
                                #     class_="mx-auto",
                                #     max_width="300",
                                #     type="card",
                                # )
                            ],
                        )

    return tabs



import ipywidgets as widgets
import playwright.sync_api
import pytest
from IPython.display import display

import solara.util

from .conftest import SERVERS


def test_widget_button_solara(solara_test, page_session: playwright.sync_api.Page, assert_solara_snapshot):
    # this all runs in process, which only works with solara
    # also, this test is only with pure ipywidgets
    button = widgets.Button(description="Click Me!")

    def change_description(obj):
        button.description = "Tested event"

    button.on_click(change_description)
    display(button)
    button_sel = page_session.locator("text=Click Me!")
    button_sel.wait_for()
    button_sel.click()
    page_session.locator("text=Tested event").wait_for()
    # with ipyvuetify 2 we modified the button font for ipywidgets, that does not happen anymore with ipyvuetify 3
    # therefore we have a different screenshot for ipyvuetify 2 and 3
    assert_solara_snapshot(page_session.locator("text=Tested event").screenshot(), postfix="-ipyvuetify3" if solara.util.ipyvuetify_major_version == 3 else "")


def test_solara_button_all(ipywidgets_runner, page_session: playwright.sync_api.Page, request, assert_solara_snapshot):
    if request.node.callspec.params["ipywidgets_runner"] != "solara" and request.node.callspec.params["solara_server"] != SERVERS[0]:
        pytest.skip("No need to run this test for all servers.")

    # this function (or rather its lines) will be executed in the kernel
    # voila, lab, classic notebook and solara will all execute it
    def kernel_code():
        import solara

        @solara.component
        def Button():
            text, set_text = solara.use_state("Click Me!")

            def on_click():
                set_text("Tested event")

            with solara.Row():
                solara.Button(text, on_click=on_click)

        display(Button())

    ipywidgets_runner(kernel_code)
    button_sel = page_session.locator("button >> text=Click Me!")
    assert_solara_snapshot(button_sel.screenshot())
    button_sel.wait_for()
    button_sel.click()
    page_session.locator("button >> text=Tested event").wait_for()
    page_session.wait_for_timeout(1000)


def test_slider_all(ipywidgets_runner, page_session: playwright.sync_api.Page, request, assert_solara_snapshot):
    if request.node.callspec.params["ipywidgets_runner"] != "solara" and request.node.callspec.params["solara_server"] != SERVERS[0]:
        pytest.skip("No need to run this test for all servers.")

    # this function (or rather its lines) will be executed in the kernel
    # voila, lab, classic notebook and solara will all execute it
    def kernel_code():
        import ipywidgets as widgets

        slider = widgets.FloatSlider(value=7.5, min=5.0, max=10.0, step=0.1, description="FloatSlider")
        box = widgets.HBox([slider])
        box.add_class("my-slider")
        # we do not want it full with, which depends on environment/browser window size
        box.layout.width = "400px"
        display(box)

    ipywidgets_runner(kernel_code)
    slider_sel = page_session.locator(".my-slider")
    assert_solara_snapshot(slider_sel.screenshot(), postfix="-ipywidgets-" + str(solara.util.ipywidgets_major))
    page_session.wait_for_timeout(1000)



import playwright.sync_api
import pytest
from IPython.display import display

from .conftest import SERVERS

# the altair figure uses the CDN


def test_cdn_via_altair(ipywidgets_runner, page_session: playwright.sync_api.Page, request, assert_solara_snapshot):
    if request.node.callspec.params["ipywidgets_runner"] != "solara" and request.node.callspec.params["solara_server"] != SERVERS[0]:
        pytest.skip("No need to run this test for all servers.")

    # this function (or rather its lines) will be executed in the kernel
    # voila, lab, classic notebook and solara will all execute it
    def kernel_code():
        import altair as alt
        from vega_datasets import data

        import solara

        source = data.seattle_weather()

        @solara.component
        def Page():
            chart = (
                alt.Chart(source, title="Daily Max Temperatures (C) in Seattle, WA")
                .mark_rect()
                .encode(
                    alt.X("date(date):O").title("Day").axis(format="%e", labelAngle=0),
                    alt.Y("month(date):O").title("Month"),
                    alt.Color("max(temp_max)").title(None),
                    tooltip=[
                        alt.Tooltip("monthdate(date)", title="Date"),
                        alt.Tooltip("max(temp_max)", title="Max Temp"),
                    ],
                )
                .configure_view(step=13, strokeWidth=0)
                .configure_axis(domain=False)
            )
            solara.AltairChart(chart)

        display(Page())

    ipywidgets_runner(kernel_code)
    vega_selector = page_session.locator('details[title="Click to view actions"]')
    vega_selector.wait_for(state="attached")
    # assert_solara_snapshot(vega_selector.screenshot())
    # page_session.wait_for_timeout(1000)



from ..utils.dataframe import df_type


def use_df_column_names(df):
    if df_type(df) == "vaex":
        return df.get_column_names()
    elif df_type(df) == "pandas":
        return df.columns.tolist()
    elif df_type(df) == "polars":
        return df.columns
    else:
        raise TypeError(f"{type(df)} not supported")



"""
# Overview
Click on one of the items on the left.
"""

import solara
from solara.alias import rv

from .. import List
from .. import SimpleListItem as ListItem

_title = "API"


# @solara.component
# def Page():
#     return solara.Markdown(__doc__)

items = [
    {
        "name": "Input",
        "icon": "mdi-chevron-left-box",
        "pages": ["button", "checkbox", "input", "select", "slider", "switch", "togglebuttons", "file_browser", "file_drop"],
    },
    {
        "name": "Output",
        "icon": "mdi-chevron-right-box",
        "pages": ["markdown", "markdown_editor", "html", "image", "sql_code", "file_download", "tooltip"],
    },
    {
        "name": "Status",
        "icon": "mdi-information",
        "pages": ["success", "info", "warning", "error", "spinner", "progress"],
    },
    {
        "name": "Viz",
        "icon": "mdi-chart-histogram",
        "pages": ["altair", "echarts", "matplotlib", "plotly", "plotly_express"],
    },
    {
        "name": "Layout",
        "icon": "mdi-page-layout-sidebar-left",
        "pages": [
            "app_layout",
            "app_bar",
            "app_bar_title",
            "card",
            "card_actions",
            "columns",
            "columns_responsive",
            "column",
            "row",
            "griddraggable",
            "gridfixed",
            "sidebar",
            "hbox",
            "vbox",
        ],
    },
    {
        "name": "Data",
        "icon": "mdi-database",
        "pages": ["dataframe", "pivot_table"],
    },
    {
        "name": "Page",
        "icon": "mdi-file-code",
        "pages": ["head", "title"],
    },
    {
        "name": "Hooks",
        "icon": "mdi-hook",
        "pages": [
            "use_cross_filter",
            "use_thread",
            "use_exception",
            "use_effect",
            "use_memo",
            "use_previous",
            "use_reactive",
            "use_state",
            "use_state_or_update",
            "use_trait_observe",
        ],
    },
    {
        "name": "Types",
        "icon": "mdi-fingerprint",
        "pages": ["route"],
    },
    {
        "name": "Routing",
        "icon": "mdi-router",
        "pages": ["use_route", "use_router", "resolve_path", "generate_routes", "generate_routes_directory", "link"],
    },
    {
        "name": "Utils",
        "icon": "mdi-hammer-wrench",
        "pages": [
            "display",
            "get_kernel_id",
            "get_session_id",
            "memoize",
            "reactive",
            "widget",
            "component_vue",
        ],
    },
    {
        "name": "Advanced",
        "icon": "mdi-head-cog-outline",
        "pages": ["style", "meta"],
    },
    {
        "name": "Cross filter",
        "icon": "mdi-filter-variant-remove",
        "pages": ["cross_filter_dataframe", "cross_filter_report", "cross_filter_slider", "cross_filter_select"],
    },
    {
        "name": "Enterprise",
        "icon": "mdi-office-building",
        "pages": ["avatar", "avatar_menu"],
    },
    {
        "name": "Lab (experimental)",
        "icon": "mdi-flask-outline",
        "pages": [
            "computed",
            "chat",
            "confirmation_dialog",
            "cookies_headers",
            "menu",
            "input_date",
            "on_kernel_start",
            "tab",
            "tabs",
            "task",
            "theming",
            "use_task",
            "use_dark_effective",
        ],
    },
]


@solara.component
def Page():
    # show a gallery of all the api pages
    router = solara.use_router()
    route_current = router.path_routes[-2]

    routes = {r.path: r for r in route_current.children}

    for item in items:
        solara.Markdown(f"## {item['name']}")
        with solara.ColumnsResponsive(12, 6, 6, 4, 4):
            for page in item["pages"]:
                if page not in routes:
                    continue
                route = routes[page]
                path = route.path
                image_url = None
                if page in [
                    "button",
                    "checkbox",
                    "confirmation_dialog",
                    "echarts",
                    "file_browser",
                    "file_download",
                    "matplotlib",
                    "select",
                    "switch",
                    "tooltip",
                ]:
                    image_url = "https://dxhl76zpt6fap.cloudfront.net/public/api/" + page + ".gif"
                elif page in ["card", "dataframe", "pivot_table", "slider"]:
                    image_url = "https://dxhl76zpt6fap.cloudfront.net/public/api/" + page + ".png"
                else:
                    image_url = "https://dxhl76zpt6fap.cloudfront.net/public/logo.svg"

                path = getattr(route.module, "redirect", path)
                if path:
                    with rv.Card(
                        elevation=2,
                        dark=False,
                        height="100%",
                        style_="display: flex; flex-direction: column; justify-content: space-between;",
                    ):
                        rv.CardTitle(children=[route.label])
                        with rv.CardText():
                            with solara.Link(path):
                                with solara.Column(align="center"):
                                    solara.Image(image_url, width="120px")
                        doc = route.module.__doc__ or ""
                        if doc:
                            lines = doc.split("\n")
                            lines = [line.strip() for line in lines if line.strip()]
                            first = lines[1]

                            rv.CardText(
                                children=[solara.Markdown(first)],
                            )


@solara.component
def NoPage():
    raise RuntimeError("This page should not be rendered")


@solara.component
def Sidebar(children=[], level=0):
    # note that we don't use children here, but we used route.module instead to ge the module
    # this is fine because all api/*.py files use the standard Page component, and do not add
    # a new Layout component
    route_current, all_routes = solara.use_route()
    if route_current is None:
        return solara.Error("Page not found")

    # keeps track of which routes we includes
    routes = {r.path: r for r in all_routes.copy()}

    def add(path):
        route = routes[path]
        with solara.Link(route):
            ListItem(route.label, class_="active" if route_current is not None and path == route_current.path else None)
        del routes[path]

    # with solara.HBox(grow=True) as main:
    with rv.Col(tag="aside", md=4, lg=3, class_="sidebar bg-grey d-none d-md-block") as main:
        with solara.Head():
            name = route_current.label if route_current.label is not None else "No name"
            if name == "API":
                solara.Title("Solara » API overview")
            else:
                solara.Title("Solara » API » " + name)
        with List():
            add("/")

            for item in items:
                with ListItem(item["name"], icon_name=item["icon"]):
                    with List():
                        for component in item["pages"]:
                            add(component)

        if routes:
            print(f"Routes not used: {list(routes.keys())}")  # noqa

    return main


@solara.component
def Layout(children=[]):
    route_current, all_routes = solara.use_route()
    if route_current is None:
        return solara.Error("Page not found")

    if route_current.path == "/":
        return Page()
    else:
        with solara.HBox(grow=True) as main:
            with solara.Padding(4):
                if route_current.module:
                    # we ignore children, and make the element again
                    WithCode(route_current.module)
        return main


@solara.component
def WithCode(module):
    # e = solara.use_exception_handler()
    # if e is not None:
    #     return solara.Error("oops")
    component = getattr(module, "Page", None)
    with rv.Sheet() as main:
        # It renders code better
        solara.Markdown(
            module.__doc__ or "# no docs yet",
            unsafe_solara_execute=True,
        )
        if component and component != NoPage:
            with solara.Card("Example", margin=0, classes=["mt-8"]):
                component()
                github_url = solara.util.github_url(module.__file__)
                solara.Button(
                    label="View source",
                    icon_name="mdi-github-circle",
                    attributes={"href": github_url, "target": "_blank"},
                    text=True,
                    outlined=True,
                    class_="mt-8",
                )
    return main



"""# MarkdownEditor
"""


import solara
from solara.alias import rv
from solara.website.utils import apidoc


@solara.component
def Page():
    markdown_initial = """
# Large heading
## Medium heading
### Small heading


## List items

   * item 1
   * item 2

## Numbered items

   1. item 1
   2. item 2


**bold** and *italic*, [GitHub link](https://github.com/widgetti/solara/)


```
preformatted code
which respecs spaces
and line breaks
```

    """.strip()

    markdown_text, set_markdown_text = solara.use_state(markdown_initial)
    with solara.HBox(grow=True) as main:
        with solara.VBox():
            with solara.Padding(2):
                with rv.Sheet(elevation=2):
                    solara.MarkdownEditor(markdown_text, on_value=set_markdown_text)
                with rv.Sheet(elevation=2):
                    solara.Markdown("# Raw markdown")
                    solara.Preformatted(markdown_text)

    return main


__doc__ += apidoc(solara.MarkdownEditor.f)  # type: ignore



"""
# DataFrame

"""

from typing import Any, Dict, Optional, cast

import plotly

import solara
from solara.website.utils import apidoc

df = plotly.data.iris()


@solara.component
def Page():
    column, set_column = solara.use_state(cast(Optional[str], None))
    cell, set_cell = solara.use_state(cast(Dict[str, Any], {}))

    def on_action_column(column):
        set_column(column)

    def on_action_cell(column, row_index):
        set_cell(dict(column=column, row_index=row_index))

    column_actions = [solara.ColumnAction(icon="mdi-sunglasses", name="User column action", on_click=on_action_column)]
    cell_actions = [solara.CellAction(icon="mdi-white-balance-sunny", name="User cell action", on_click=on_action_cell)]
    solara.Markdown(
        f"""
        ## Demo

        Below we show display the titanic dataset, and demonstrate a user column and cell action. Try clicking on the triple icon when hovering
        above a column or cell. And see the following values changes:

        * Column action on: `{column}`
        * Cell action on: `{cell}`

    """
    )
    solara.DataFrame(df, column_actions=column_actions, cell_actions=cell_actions)


__doc__ += apidoc(solara.DataFrame.f)  # type: ignore



import contextlib

import solara
from solara.alias import rv
from solara.components.title import Title

from ..components import Header, Hero
from ..components.mailchimp import MailChimp

title = "Home"

route_order = ["/", "showcase", "docs", "api", "examples", "apps"]


@solara.component
def Page():
    solara.Markdown("should not see me")


@solara.component
def SimpleListItem(text=None, children=[], class_: str = None, icon_name=None):
    if icon_name is not None:
        children = [rv.Icon(children=[icon_name])] + children
    if text is not None:
        children = [text] + children
    return rv.Html(tag="li", children=children, attributes={"class": class_})


@solara.component
def List(children=[], class_: str = None):
    return rv.Html(tag="ul", children=children, attributes={"class": class_})


@solara.component
def Sidebar():
    route_current, all_routes = solara.use_route()
    router = solara.use_router()
    selected = router.path
    with rv.Col(tag="aside", md=4, lg=3, class_="sidebar bg-grey d-none d-md-block") as main:
        with List():
            for route in all_routes:
                if route.children and route.data is None:
                    path = solara.resolve_path(route.children[0])
                    path = getattr(route.module, "redirect", path)
                    path = getattr(route.children[0].module, "redirect", path)
                    with solara.Link(path) if path is not None else contextlib.nullcontext():
                        with SimpleListItem(route.label, class_="active" if path == selected else None):
                            with List():
                                for child in route.children[1:]:
                                    path = solara.resolve_path(child)
                                    path = getattr(child.module, "redirect", path)
                                    with solara.Link(path) if path is not None else contextlib.nullcontext():
                                        title = child.label or "no label"
                                        if callable(title):
                                            title = "Error: dynamic title"
                                        SimpleListItem(title, class_="active" if path == selected else None)
                else:
                    path = solara.resolve_path(route)
                    path = getattr(route.module, "redirect", path)
                    if route.children:
                        path = getattr(route.children[0].module, "redirect", path)
                    with solara.Link(path) if path is not None else contextlib.nullcontext():
                        SimpleListItem(route.label, class_="active" if path == selected else None)
    return main


@solara.component
def Layout(children=[]):
    router = solara.use_router()
    route_current, all_routes = solara.use_route()
    route_sidebar_current, all_routes_sidebar = solara.use_route(1)

    show_left_menu, set_show_left_menu = solara.use_state(False)
    show_right_menu, set_show_right_menu = solara.use_state(False)

    target, set_target = solara.use_state(0)

    if route_current and route_current.path == "apps":
        return children[0]
    with solara.VBox(grow=False) as main:
        Title(title="Solara documentation")
        solara.Meta(name="twitter:card", content="summary_large_image")
        solara.Meta(name="twitter:site", content="@solara_dev")
        solara.Meta(name="twitter:image", content="https://solara.dev/static/assets/images/logo-small.png")
        solara.Meta(property="og:url", content="https://solara.dev" + router.path)
        solara.Meta(property="og:image", content="https://solara.dev/static/assets/images/logo-small.png")
        solara.Meta(property="og:type", content="website")
        Header(
            on_toggle_left_menu=lambda: set_show_left_menu(not show_left_menu),
            on_toggle_right_menu=lambda: set_show_right_menu(not show_right_menu),
        )
        if route_current is not None and route_current.path == "/":
            Hero(
                title="A pure Python, React-style web framework",
                sub_title="Solara helps you build powerful & scalable Jupyter and web apps <b>faster</b> and <b>easier</b>.",
                button_text="Quickstart",
            )

        with rv.Container(tag="section", fluid=True, ma_0=True, pa_0=True, class_="fill-height mb-8 solara-content-main"):
            if route_current is not None and route_current.path == "/":
                description = "Use ipywidgets with Solara to build powerful and scalable web apps for Jupyter and production in Python."
                # both tags in one
                solara.Meta(name="description", property="og:description", content=description)
                solara.Meta(name="twitter:description", content=description)
                solara.Meta(property="og:title", content="Solara documentation")
                solara.Meta(name="twitter:title", content="Solara documentation")

                with rv.Row(class_="ma-2"):
                    with rv.Col(md=6, offset_md=3, sm=8, offset_sm=2):
                        solara.Markdown(
                            """
                        # What is Solara?

                        Solara lets you build web apps from pure Python using ipywidgets or a React-like API on
                        top of ipywidgets.
                        These apps work both inside the Jupyter Notebook and as standalone web apps with frameworks like FastAPI.

                        With Solara, you benefit from a paradigm that promotes component-based code and simplifies state management,
                        making your development process more efficient and your applications more maintainable.

                        Solara provides you with access to the full strength of the Python ecosystem.
                        This means you can continue using your favorite libraries while expanding your web development capabilities.
                        """
                        )
                        with solara.HBox():
                            with solara.Link("/docs"):
                                solara.Button(label="Read more", class_="ma-1 homepage-button", href="/docs", color="primary", dark=True)
                            with solara.Link("/docs/quickstart"):
                                solara.Button(label="Quickstart", class_="ma-1 homepage-button", color="primary", dark=True)
                    # with rv.Col(md=4, sm=5):
                    #     rv.Img(src="https://dxhl76zpt6fap.cloudfront.net/public/landing/what.webp", style_="width:900px")

                with solara.Column(style={"width": "100%"}, gap="2.5em", classes=["pt-10", "mt-8"], align="center"):
                    with solara.Row(justify="center", gap="2.5em", classes=["ma-2", "row-container"]):
                        rv.Img(src="https://dxhl76zpt6fap.cloudfront.net/public/landing/complexity.webp", contain=True)
                        solara.Markdown(
                            """
                        # Build **large** apps with **low** code complexity

                        With Solara, you can build large scale apps without hitting a complexity wall.

                        With other tools you may hit a dead end due to missing features or implementing features
                        adds too much complexity to your code base.

                        Solara offers the **flexibility** to build complex apps, but keeps the **simplicity** of a small code base.
                        """
                        )
                    with solara.Row(justify="center", gap="2.5em", classes=["ma-2", "row-container"]):
                        solara.Markdown(
                            """
                        # The trustworthiness of React

                        Using the same API as React, but ported to Python, Solara lets you build apps with the same
                        trustworthiness as React.

                        With a decade of experience, React is battle-tested and proven to be
                        a reliable and robust framework to build large scale apps.
                        """
                        )
                        with solara.Row(justify="center", style={"width": "500px"}):
                            rv.Img(src="https://dxhl76zpt6fap.cloudfront.net/public/landing/python-love-react.webp", style_="max-width:300px", contain=True)
                    with solara.Row(justify="center", gap="2.5em", classes=["ma-2", "row-container"]):
                        with solara.Column():
                            if target == 0:
                                solara.Markdown("#### Running in: Jupyter notebook")
                                solara.Image(
                                    "https://global.discourse-cdn.com/standard11/uploads/jupyter/original/2X/8/8bc875c0c3845ae077168575a4f8a49cf1b35bc6.gif"
                                )
                            else:
                                solara.Markdown("#### Running in: FastAPI")
                                solara.Image(
                                    "https://global.discourse-cdn.com/standard11/uploads/jupyter/original/2X/9/9442fc70e2a1fcd201f4f900fa073698a1f8c937.gif"
                                )
                            import solara.website.pages.apps.scatter as scatter

                            github_url = solara.util.github_url(scatter.__file__)
                            solara.Button(
                                label="View source",
                                icon_name="mdi-github-circle",
                                attributes={"href": github_url, "target": "_blank"},
                                text=True,
                                outlined=False,
                            )
                            with solara.Link("/examples"):
                                with solara.Column(style="width: 100%;"):
                                    solara.Button(
                                        label="More examples",
                                        icon_name="mdi-brain",
                                        text=True,
                                        outlined=False,
                                    )
                        with solara.Column():
                            solara.Markdown(
                                """
                                    ## Create apps

                                    In Jupyter or standalone, and run them in production
                                    using FastAPI or starlette.

                                    Get more inspiration from our [examples](/examples).
                                """
                            )
                            with rv.ExpansionPanels(v_model=target, on_v_model=set_target, mandatory=True, flat=True):
                                with rv.ExpansionPanel():
                                    rv.ExpansionPanelHeader(children=["Jupyter notebook"])
                                    with rv.ExpansionPanelContent():
                                        solara.Markdown("Build on top of ipywidgets, solara components work in all Jupyter notebook environments.")
                                with rv.ExpansionPanel():
                                    rv.ExpansionPanelHeader(children=["FastAPI"])
                                    with rv.ExpansionPanelContent():
                                        solara.Markdown(
                                            "Using [solara-server](/docs/understanding/solara-server), we can run our app in production using FastAPI."
                                        )

                with solara.Column(style={"width": "100%"}):
                    solara.v.Divider()

                with solara.Column(align="center", gap="2.5em", style={"width": "100%", "padding-bottom": "50px"}):
                    solara.Markdown("# Testimonials", style="text-align:center")
                    with solara.Row(justify="center", gap="2.5em", style={"align-items": "stretch", "flex-wrap": "wrap", "row-gap": "2.5em"}):
                        Testimonial(
                            "Solara is like streamlit, but for Jupyter. I am really excited to see where this goes!",
                            "Jack Parmer",
                            "Former CEO and Co-Founder of Plotly",
                            "https://dxhl76zpt6fap.cloudfront.net/public/avatar/jack-parmer.jpg",
                        )
                        Testimonial(
                            "Solara has been transformative, allowing us to rapidly create a Jupyter app and iterate with impressive speed.",
                            "Nick Elprin",
                            "CEO and Co-Founder of Domino Data Lab",
                            "https://dxhl76zpt6fap.cloudfront.net/public/avatar/nick-elprin.jpg",
                        )
                        Testimonial(
                            "Solara allows us to go from prototype to production with the same stack.",
                            "Jonathan Chambers",
                            "Co-founder of Planeto",
                            "https://dxhl76zpt6fap.cloudfront.net/public/avatar/jonathan-chambers.jpg",
                        )

                with solara.Column(style={"width": "100%"}):
                    solara.v.Divider()

                with solara.Column(align="center", gap="2.5em", style={"width": "100%", "padding-bottom": "50px"}):
                    solara.Markdown("# Sponsors", style="text-align:center")
                    with solara.Row(justify="center", gap="2.5em", style={"align-items": "stretch"}):
                        with solara.v.Html(tag="a", attributes={"href": "https://www.dominodatalab.com/", "target": "_blank"}):
                            solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/sponsors/domino.png", width="300px")

                with solara.Column(style={"width": "100%"}):
                    solara.v.Divider()

                with solara.Row(justify="center", gap="2.5em", classes=["footer-wrapper"]):
                    with solara.Column(align="center", style={"min-width": "300px"}):
                        solara.Markdown(
                            """
                            #### For any consulting, training or support needs
                            [contact@solara.dev](mailto:contact@solara.dev)
                            """
                        )
                    solara.v.Divider(vertical=True)
                    with solara.Column(align="center", style={"min-width": "300px"}):
                        solara.Markdown("#### Join our Mailing list to get the latest news")
                        with solara.Div(style={"width": "80%"}):
                            MailChimp(location=router.path)

            else:
                with rv.Row(
                    style_="gap:6rem; flex-wrap: nowrap;", justify="center" if route_current is not None and route_current.path == "showcase" else "start"
                ):
                    if route_current is not None and hasattr(route_current.module, "Sidebar"):
                        route_current.module.Sidebar()  # type: ignore
                    else:
                        if route_current is not None and route_current.path != "showcase":
                            Sidebar()
                    with rv.Col(tag="main", md=True, class_="pt-12 pl-12 pr-10", style_="max-width: 1024px; overflow: auto;"):
                        if route_current is not None and route_current.path == "/":
                            with rv.Row(align="center"):
                                pass
                                # with rv.Col(md=6, class_="pa-0"):
                                #     rv.Html(tag="h1", children=["Live Demo"])
                                # with rv.Col(md=6, class_="d-flex", style_="justify-content: end"):
                                #     rv.Btn(elevation=0, large=True, children=["Running App"], color="primary", class_="btn-size--xlarge")
                            # solara.Padding(6)
                        with rv.Row(children=children, class_="solara-page-content-search"):
                            pass

            # absolute = True prevents the drawer from being below the overlay it generates
            # Drawer navigation for top menu
            with rv.NavigationDrawer(
                v_model=show_right_menu,
                on_v_model=set_show_right_menu,
                fixed=True,
                absolute=True,
                right=True,
                hide_overlay=False,
                overlay_color="#000000",
                overlay_opacity=0.5,
                style_="height: 100vh",
            ):
                with rv.List(nav=True):
                    with rv.ListItemGroup(active_class="text--primary"):
                        for route in all_routes:
                            if route.path == "apps":
                                continue
                            with solara.Link(route):
                                solara.ListItem(route.label)

            # Drawer navigation for sidebar
            with rv.NavigationDrawer(
                v_model=show_left_menu,
                on_v_model=set_show_left_menu,
                fixed=True,
                absolute=True,
                hide_overlay=False,
                overlay_color="#000000",
                overlay_opacity=0.5,
                style_="height: 100vh",
            ):
                with rv.List(nav=True):
                    current_path = router.path
                    with rv.ListItemGroup(active_class="text--primary", v_model=current_path):
                        for route in all_routes_sidebar:
                            # this gets rid of the api/link child routes
                            if len([k for k in route.children if k.label]) == 0:
                                with solara.Link(route):
                                    solara.ListItem(route.label, value=solara.resolve_path(route))
                            else:
                                with solara.ListItem(route.label or "no-title", value=solara.resolve_path(route) + "_no_select"):
                                    for subroute in route.children:
                                        if subroute.label:
                                            with solara.Link(subroute):
                                                with solara.ListItem(subroute.label, value=solara.resolve_path(subroute)):
                                                    pass

    return main


@solara.component
def Testimonial(text, name, position, img):
    max_width = "300px"
    with rv.Card(
        elevation=2,
        dark=False,
        max_width=max_width,
        class_="testimonial-card",
    ):
        # rv.CardTitle(children=["Former Plotly CEO"])
        with rv.CardActions():
            with rv.ListItem(class_="grow"):
                with rv.ListItemAvatar(color="grey darken-3"):
                    rv.Img(
                        class_="elevation-6",
                        src=img,
                    )
                with rv.ListItemContent():
                    rv.ListItemTitle(children=[name])
                    rv.ListItemSubtitle(children=[position], style_="white-space: unset;")
        rv.CardText(
            children=[text],
            style_="font-style: italic; padding-top: 0px",
        )



"""
# HBox
Lays out children in horizontal direction.
"""
import solara

from .common import ColorCard


@solara.component
def Page():
    with solara.VBox() as main:
        colors = "green red orange brown yellow pink".split()
        with solara.HBox():
            for color in colors:
                ColorCard(color, color)
    return main



"""
# AppBarTitle

"""
import solara
import solara.lab
from solara.website.utils import apidoc

from . import NoPage

title = "AppBarTitle"

Page = NoPage


__doc__ += apidoc(solara.AppBarTitle.f)  # type: ignore



import json
from pathlib import Path
from typing import List

# based on notebook config, but to reduce the dependencies
# and with modern typing and use of pathlib


def recursive_update(target, new):
    """Recursively update one dictionary using another.

    None values will delete their keys.
    """
    for k, v in new.items():
        if isinstance(v, dict):
            if k not in target:
                target[k] = {}
            recursive_update(target[k], v)
            if not target[k]:
                # Prune empty subdicts
                del target[k]

        elif v is None:
            target.pop(k, None)

        else:
            target[k] = v


def _get_config(directory: Path, config_name: str):
    paths = [directory / f"{config_name}.json"]
    paths.extend((directory / f"{config_name}.d/").glob("*.json"))
    data: dict = {}
    for path in paths:
        if path.is_file():
            with open(path, encoding="utf-8") as f:
                recursive_update(data, json.load(f))
    return data


def get_config(directories: List[Path], config_name: str):
    config: dict = {}
    # step through back to front, to ensure front of the list is top priority
    for directory in directories[::-1]:
        recursive_update(config, _get_config(directory, config_name))
    return config



"""

Wraps plotly express function and adds cross filtering support.

Instead of `plotly.express' you can use `solara.express` instead.

```python
# import plotly.express as px
import solara.express as px

df = px.data.iris()
px.histogram(df, "species")
px.scatter(df, x="sepal_width", y="sepal_length", color="species")
```


Click the lasso icon in the top scatter plot to select points, which should then be filtered out in the other
plots.


"""

import plotly.express as px

import solara
import solara.express as spx

df = px.data.iris()


@solara.component
def Page():
    solara.provide_cross_filter()
    fig = px.histogram(df, "species")
    fig.update_layout(dragmode="select", selectdirection="h")

    with solara.VBox() as main:
        spx.scatter(df, x="sepal_width", y="sepal_length", color="species")
        spx.scatter_3d(df, x="sepal_width", y="sepal_length", z="petal_width")
        spx.CrossFilteredFigurePlotly(fig)
    return main



import abc
import asyncio
import dataclasses
import inspect
import logging
import threading
from enum import Enum
from typing import (
    Any,
    Callable,
    Coroutine,
    Generic,
    List,
    Optional,
    TypeVar,
    Union,
    cast,
    overload,
)

import typing_extensions

import solara
import solara.util
from solara.toestand import Singleton

from .toestand import Ref as ref

R = TypeVar("R")
T = TypeVar("T")
P = typing_extensions.ParamSpec("P")

logger = logging.getLogger("solara.task")

try:
    threading.Thread(target=lambda: None).start()
    has_threads = True
except RuntimeError:
    has_threads = False
has_threads


class TaskState(Enum):
    NOTCALLED = 1
    STARTING = 2
    WAITING = 3
    RUNNING = 4
    ERROR = 5
    FINISHED = 6
    CANCELLED = 7


@dataclasses.dataclass(frozen=True)
class TaskResult(Generic[T]):
    value: Optional[T] = None
    latest: Optional[T] = None
    exception: Optional[Exception] = None
    # only useful if you want to know details about the state like STARTING or WAITING
    _state: TaskState = TaskState.NOTCALLED
    progress: Optional[float] = None

    @property
    def not_called(self):
        return self._state == TaskState.NOTCALLED

    @property
    def pending(self):
        return self._state in (TaskState.STARTING, TaskState.WAITING, TaskState.RUNNING)

    @property
    def finished(self):
        return self._state == TaskState.FINISHED

    @property
    def cancelled(self):
        return self._state == TaskState.CANCELLED

    @property
    def error(self):
        return self._state == TaskState.ERROR


class Task(Generic[P, R], abc.ABC):
    def __init__(self):
        self._result = solara.reactive(
            TaskResult[R](
                value=None,
                _state=TaskState.NOTCALLED,
            )
        )
        self._last_value: Optional[R] = None
        self._last_progress: Optional[float] = None
        self._latest = ref(self._result.fields.latest)
        self._value = ref(self._result.fields.value)
        self._error = ref(self._result.fields.error)
        self._finished = ref(self._result.fields.finished)
        self._cancelled = ref(self._result.fields.cancelled)
        self._pending = ref(self._result.fields.pending)
        self._not_called = ref(self._result.fields.not_called)
        self._progress = ref(self._result.fields.progress)
        self._exception = ref(self._result.fields.exception)
        self._state_ = ref(self._result.fields._state)

    @property
    def result(self) -> TaskResult[R]:
        return self._result.value

    @property
    def latest(self) -> Optional[R]:
        return self._latest.value

    @property
    def value(self) -> Optional[R]:
        return self._value.value

    @property
    def _state(self) -> TaskState:
        return self._state_.value

    @property
    def error(self) -> bool:
        return self._error.value

    @property
    def finished(self) -> bool:
        return self._finished.value

    @property
    def cancelled(self) -> bool:
        return self._cancelled.value

    @property
    def pending(self) -> bool:
        return self._pending.value

    @property
    def not_called(self) -> bool:
        return self._not_called.value

    @property
    def progress(self) -> Optional[float]:
        return self._progress.value

    @progress.setter
    def progress(self, value: Optional[float]) -> None:
        self._last_progress = value
        self._progress.value = value

    @property
    def exception(self) -> Optional[Exception]:
        return self._exception.value

    @abc.abstractmethod
    def retry(self) -> None:
        ...

    @abc.abstractmethod
    def cancel(self) -> None:
        ...

    @abc.abstractmethod
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        ...

    @abc.abstractmethod
    def is_current(self) -> bool:
        ...

    def _prestart(self):
        self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.STARTING)


class _CancelledErrorInOurTask(BaseException):
    pass


class TaskAsyncio(Task[P, R]):
    current_task: Optional[asyncio.Task] = None
    current_future: Optional[asyncio.Future] = None
    _cancel: Optional[Callable[[], None]] = None
    _retry: Optional[Callable[[], None]] = None

    def __init__(self, run_in_thread: bool, function: Callable[P, Coroutine[Any, Any, R]]):
        self.run_in_thread = run_in_thread
        self.function = function
        super().__init__()

    def cancel(self) -> None:
        if self._cancel:
            self._cancel()
        else:
            raise RuntimeError("Cannot cancel task, never started")

    def retry(self):
        if self._retry:
            self._retry()
        else:
            raise RuntimeError("Cannot retry task, never started")

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        self.current_future = future = asyncio.Future[R]()
        self._last_progress = None
        current_task: asyncio.Task[None]
        if self.current_task:
            self.current_task.cancel()

        def retry():
            self(*args, **kwargs)

        def cancel():
            event_loop = current_task.get_loop()
            # cancel after cancel is a no-op
            self._cancel = lambda: None
            if asyncio.current_task() == current_task:
                if event_loop == asyncio.get_event_loop():
                    # we got called in our own task and event loop
                    raise _CancelledErrorInOurTask()
                else:
                    current_task.cancel()
                    self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.CANCELLED)
            else:
                current_task.cancel()
                self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.CANCELLED)

        self._cancel = cancel
        self._retry = retry
        call_event_loop = asyncio.get_event_loop()

        if self.run_in_thread:
            thread_event_loop = asyncio.new_event_loop()
            self.current_task = current_task = thread_event_loop.create_task(self._async_run(call_event_loop, future, args, kwargs))

            def runs_in_thread():
                try:
                    thread_event_loop.run_until_complete(current_task)
                except asyncio.CancelledError as e:
                    call_event_loop.call_soon_threadsafe(future.set_exception, e)
                except Exception as e:
                    logger.exception("error running in thread")
                    call_event_loop.call_soon_threadsafe(future.set_exception, e)
                    raise

            self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.STARTING)
            thread = threading.Thread(target=runs_in_thread)
            thread.start()
        else:
            self.current_task = current_task = asyncio.create_task(self._async_run(call_event_loop, future, args, kwargs))
            self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.STARTING)

    def is_current(self):
        running_task = self.current_task
        assert running_task is not None
        return (self.current_task == asyncio.current_task()) and not running_task.cancelled()

    async def _async_run(self, call_event_loop: asyncio.AbstractEventLoop, future: asyncio.Future, args, kwargs) -> None:
        task_for_this_call = asyncio.current_task()
        assert task_for_this_call is not None

        if self.is_current():
            self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.STARTING)

        async def runner():
            try:
                if self.is_current():
                    self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.RUNNING)
                self._last_value = value = await self.function(*args, **kwargs)
                if self.is_current() and not task_for_this_call.cancelled():  # type: ignore
                    self._result.value = TaskResult[R](value=value, latest=value, _state=TaskState.FINISHED, progress=self._last_progress)
                logger.info("setting result to %r", value)
                call_event_loop.call_soon_threadsafe(future.set_result, value)
            except Exception as e:
                if self.is_current():
                    logger.exception(e)
                    self._result.value = TaskResult[R](latest=self._last_value, exception=e, _state=TaskState.ERROR)
                call_event_loop.call_soon_threadsafe(future.set_exception, e)
            # Although this seems like an easy way to handle cancellation, an early cancelled task will never execute
            # so this code will never execute, so we need to handle this in the cancel function in __call__
            # except asyncio.CancelledError as e:
            #    if self.is_current():
            #        self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.CANCELLED)
            #    call_event_loop.call_soon_threadsafe(future.set_exception, e)
            # But... if we call cancel in our own task, we still need to do it from this place
            except _CancelledErrorInOurTask as e:
                try:
                    # maybe there is a different way to get a full stack trace?
                    raise asyncio.CancelledError() from e
                except asyncio.CancelledError as e:
                    if self.is_current():
                        self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.CANCELLED)
                    call_event_loop.call_soon_threadsafe(future.set_exception, e)

        await runner()


class TaskThreaded(Task[P, R]):
    _current_cancel_event: Optional[threading.Event] = None
    _current_thread: Optional[threading.Thread] = None
    _last_finished_event: Optional[threading.Event] = None
    _cancel: Optional[Callable[[], None]] = None
    _retry: Optional[Callable[[], None]] = None

    def __init__(self, function: Callable[P, R]):
        super().__init__()
        self.__qualname__ = function.__qualname__
        self.function = function
        self.lock = threading.Lock()

    def cancel(self) -> None:
        if self._cancel:
            self._cancel()
        else:
            raise RuntimeError("Cannot cancel task, never started")

    def retry(self):
        if self._retry:
            self._retry()
        else:
            raise RuntimeError("Cannot retry task, never started")

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        self._last_finished_event = _last_finished_event = threading.Event()
        self._current_cancel_event = cancel_event = threading.Event()
        self._last_progress = None

        def retry():
            self(*args, **kwargs)

        def cancel():
            cancel_event.set()
            if threading.current_thread() == current_thread:
                raise solara.util.CancelledError()
            self._current_cancel_event = None

        self._retry = retry
        self._cancel = cancel

        with self.lock:
            previous_thread = self._current_thread
            self._current_thread = current_thread = threading.Thread(
                target=lambda: self._run(_last_finished_event, previous_thread, cancel_event, args, kwargs), daemon=False
            )
        self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.STARTING)
        current_thread.start()

    def is_current(self):
        return self._current_thread == threading.current_thread()

    def _run(self, _last_finished_event, previous_thread: Optional[threading.Thread], cancel_event, args, kwargs) -> None:
        # use_thread has this as default, which can make code run 10x slower
        intrusive_cancel = False
        wait_on_previous = False

        def runner():
            if wait_on_previous:
                if previous_thread and previous_thread.is_alive():
                    if self.is_current():
                        self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.WAITING)
                    # don't start before the previous is stopped
                    try:
                        previous_thread.join()
                    except:  # noqa
                        pass
            if self.is_current():
                self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.RUNNING)
            else:
                # early stop
                return

            callback = self.function
            try:
                guard = solara.util.cancel_guard(cancel_event) if intrusive_cancel else solara.util.nullcontext()
                try:
                    # we only use the cancel_guard context manager around
                    # the function calls to f. We don't want to guard around
                    # a call to react, since that might slow down rendering
                    # during rendering
                    with guard:
                        if self.is_current():
                            value = callback(*args, **kwargs)
                    if inspect.isgenerator(value):
                        generator = value
                        self._last_value = None
                        while True:
                            try:
                                with guard:
                                    self._last_value = value = next(generator)
                                    if self.is_current():
                                        self._result.value = TaskResult[R](latest=value, value=value, _state=TaskState.RUNNING, progress=self._last_progress)
                            except StopIteration:
                                break
                        if self.is_current():
                            self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.FINISHED, progress=self._last_progress)
                    else:
                        self._last_value = value
                        if self.is_current():
                            self._result.value = TaskResult[R](latest=value, value=value, _state=TaskState.FINISHED, progress=self._last_progress)
                except Exception as e:
                    if self.is_current():
                        logger.exception(e)
                        self._last_value = None
                        self._result.value = TaskResult[R](latest=self._last_value, exception=e, _state=TaskState.ERROR)
                    return
                except solara.util.CancelledError:
                    pass
                    # this means this thread is cancelled not be request, but because
                    # a new thread is running, we can ignore this
            finally:
                if self.is_current():
                    self.running_thread = None
                    logger.info("thread done!")
                    if cancel_event.is_set():
                        self._result.value = TaskResult[R](latest=self._last_value, _state=TaskState.CANCELLED)
                _last_finished_event.set()

        try:
            runner()
        except Exception:
            logger.exception("error running in thread")
            raise


# TODO: Not sure if we want to use this, or have all local variables in Task subclasses be reactive vars
class Proxy:
    def __init__(self, factory):
        self._instance = Singleton(factory)

    def __getattr__(self, name):
        return getattr(self._instance.value, name)

    def __setattr__(self, name, value):
        if name == "_instance":
            super().__setattr__(name, value)
        else:
            setattr(self._instance.value, name, value)

    def __call__(self, *args, **kwargs):
        return self._instance.value(*args, **kwargs)


@overload
def task(
    f: None = None,
    *,
    prefer_threaded: bool = ...,
) -> Callable[[Callable[P, R]], Task[P, R]]:
    ...


@overload
def task(
    f: Callable[P, Union[Coroutine[Any, Any, R], R]],
    *,
    prefer_threaded: bool = ...,
) -> Task[P, R]:
    ...


def task(
    f: Union[None, Callable[P, Union[Coroutine[Any, Any, R], R]]] = None,
    *,
    prefer_threaded: bool = True,
) -> Union[Callable[[Callable[P, R]], Task[P, R]], Task[P, R]]:
    """Decorator to turn a function or coroutine function into a task.

    Lets you run code in the background, with the UI available to the user. This is useful for long running tasks, like downloading data or processing data.

    The task decorator turns a function or coroutine function (`async def foo(...)` - here foo is called a coroutine function) into a task object.
    A task is a callable that will run the function or coroutine function in a separate thread
    Note that on platforms where threads are supported, asyncio tasks will still be executed in threads (unless the
    `prefer_thread=False` argument is passed). Because a coroutine function might still call long running blocking code.
    Running the asyncio task in a thread will still result in a responsive UI when executed in a separate thread.

    The task object will execute the function only once per virtual kernel and will only store one result per virtual kernel.
    When called multiple times, the previously started thread or asyncio task result will be ignored.

    A running thread or asyncio task can check if it is still the current task by calling `task.is_current()`.
    If `task.is_current()` returns False, the task should stop running and return early.

    The return value of the function is available as the `.value` reactive property on the task object, meaning that if a
    component accesses it, the component will automatically re-run when the value changes, like a [reactive variable](/api/reactive).

    ## Task object

    The task object has the following attributes/values which are all reactive:

    * `.value`: Contains the return value of the function (Only valid if `.finished` is true, else None).
    * `.exception`: The exception raised by the function, if any (Only valid if `.error` is true, else None).
    * `.latest` The last return value of the function, useful for showing out-of-date data while the task is running.
    * `.progress` A readable and writable reactive property which can be used for communicating progress to the user.

    The state of the task can be queried with the following attributes, which are all reactive:

    * `.not_called`: True if the task has not been called yet.
    * `.pending`: True if the task is asked to run, but did not finish yet, did not error and did not get cancelled.
        When true, often a loading or busy indicator is shown to the user.
    * `.finished`: True if the task has finished running. The result is available in the `.value` attribute as
        well as the `.latest` attribute.
    * `.cancelled`: True if the task was cancelled (by calling `.cancel()`).
    * `.error`: True if the function has raised an exception.

    The following methods are available:

    * `(*args, **kwargs)` : Call the task with the given arguments and keyword arguments. The task will only run once per virtual kernel.
    * `.cancel()`: Cancels the task.
    * `is_current()`: Returns True if the task is still the current task, and should continue running.
        Will return False when a new call to the task is made, and this function is being called from the the
        previous thread or asyncio.

    ## State diagram

    The following state diagram shows the possible states of a task and how
    each state transitions to another state.

    ```mermaid
    stateDiagram-v2
        not_called --> pending:  task()
        pending --> finished
        pending --> error: exception
        pending --> pending:  task()
        pending --> cancelled: task.cancel()
        finished --> pending: task()
        error --> pending: task()
        cancelled --> pending: task()
    ```

    Note that calling the task (as indicated by `task()`) can be done from any state.


    ## Example

    ### Async task


    ```solara
    import asyncio
    import solara
    from solara.lab import task

    @task
    async def fetch_data():
        await asyncio.sleep(2)
        return "The answer is 42"

    @solara.component
    def Page():
        solara.Button("Fetch data", on_click=fetch_data)
        solara.ProgressLinear(fetch_data.pending)

        if fetch_data.finished:
            solara.Text(fetch_data.value)
        elif fetch_data.not_called:
            solara.Text("Click the button to fetch data")
        # Optional state check
        # elif fetch_data.cancelled:
        #     solara.Text("Cancelled the fetch")
        # elif fetch_data.error:
        #     solara.Error(str(fetch_data.exception))

    ```

    ### Threaded task

    ```solara
    import time
    import solara
    from solara.lab import task

    @task
    def fetch_data():
        time.sleep(2)
        return "The answer is 42"


    @solara.component
    def Page():
        solara.Button("Fetch data", on_click=fetch_data)
        solara.ProgressLinear(fetch_data.pending)

        if fetch_data.finished:
            solara.Text(fetch_data.value)
        elif fetch_data.not_called:
            solara.Text("Click the button to fetch data")
        # Optional state check
        # elif fetch_data.cancelled:
        #     solara.Text("Cancelled the fetch")
        # elif fetch_data.error:
        #    solara.Error(str(fetch_data.exception))
    ```

    Note that both examples are very similar. In the first example however, we wrap a coroutine function
    which can use `asyncio.sleep`. In the second example, we use a regular function, which uses `time.sleep`.
    If the coroutine function would use `time.sleep` in combination with `prefer_threaded=False`,
    the UI would be unresponsive for 2 seconds.


    ### Showing a progress bar


    Using the `.progress` attribute, you can show a progress bar to the user. This is useful for long running tasks
    but requires a bit more work.

    ```solara
    import time
    import solara
    from solara.lab import task


    @task
    def my_calculation():
        total = 0
        for i in range(10):
            my_calculation.progress = (i + 1) * 10.0
            time.sleep(0.4)
            if not my_calculation.is_current():
                # a new call was made before this call was finished
                return
            total += i**2
        return total


    @solara.component
    def Page():
        solara.Button("Run calculation", on_click=my_calculation)
        solara.ProgressLinear(my_calculation.progress if my_calculation.pending else False)

        if my_calculation.finished:
            solara.Text(f"Calculation result: {my_calculation.value}")
        elif my_calculation.not_called:
            solara.Text("Click the button to fetch data")
        # Optional state check
        # elif my_calculation.cancelled:
        #     solara.Text("Cancelled the fetch")
        # elif my_calculation.error:
        #    solara.Error(str(my_calculation.exception))
    ```

    ### Out-of-date data

    ```solara
    import time
    import solara
    from solara.lab import task


    @task
    def my_calculation():
        total = 0
        for i in range(10):
            time.sleep(0.1)
            total += i**2
        return total


    @solara.component
    def Page():
        solara.ProgressLinear(my_calculation.pending)
        solara.Button("Run simulation", on_click=my_calculation)
        print(my_calculation.pending, my_calculation.value)

        if my_calculation.finished:
            solara.Text(f"Simulation result: {my_calculation.value}")
        if my_calculation.pending and my_calculation.latest:
            solara.Text(f"Simulation previous result: {my_calculation.latest}", style={"opacity": ".3"})
        elif my_calculation.not_called:
            solara.Text("Click the button to fetch data")
    ```

    ## Arguments

    - `f`: Function to turn into task or None
    - `prefer_threaded` - bool: Will run coroutine functions as a task in a thread when threads are available.
        This ensures that even when a coroutine functions calls a blocking function the UI is still responsive.
        On platform where threads are not supported (like Pyodide / WASM / Emscripten / PyScript), a coroutine
        function will always run in the current event loop.

    ```

    """

    def wrapper(f: Union[None, Callable[P, Union[Coroutine[Any, Any, R], R]]]) -> Task[P, R]:
        def create_task():
            if inspect.iscoroutinefunction(f):
                return TaskAsyncio[P, R](prefer_threaded and has_threads, f)
            else:
                return TaskThreaded[P, R](cast(Callable[P, R], f))

        return cast(Task[P, R], Proxy(create_task))

    if f is None:
        return wrapper
    else:
        return wrapper(f)


@overload
def use_task(
    f: None = None,
    *,
    dependencies: None = ...,
    raise_error=...,
    prefer_threaded=...,
) -> Callable[[Callable[P, R]], Task[P, R]]:
    ...


@overload
def use_task(
    f: Callable[P, R],
    *,
    dependencies: None = ...,
    raise_error=...,
    prefer_threaded=...,
) -> Task[P, R]:
    ...


@overload
def use_task(
    f: None = None,
    *,
    dependencies: List = ...,
    raise_error=...,
    prefer_threaded=...,
) -> Callable[[Callable[[], R]], "Task[[], R]"]:
    ...


@overload
def use_task(
    f: Callable[[], R],
    *,
    dependencies: List = ...,
    raise_error=...,
    prefer_threaded=...,
) -> "Task[[], R]":
    ...


def use_task(
    f: Union[None, Callable[P, R]] = None,
    *,
    dependencies: Union[None, List] = [],
    raise_error=True,
    prefer_threaded=True,
) -> Union[Callable[[Callable[P, R]], Task[P, R]], Task[P, R]]:
    """A hook that runs a function or coroutine function as a task and returns the result.

    Allows you to run code in the background, with the UI available to the user. This is useful for long running tasks,
    like downloading data or processing data.

    Unlike with the [`@task`](/api/task) decorator, the result is not globally shared, but only available to the component that called `use_task`.

    Note that unlike the [`@task`](/api/task) decorator, the task is invoked immediately when dependencies are passed.


    ## Example

    ### Running in a thread

    ```solara
    import time
    import solara
    from solara.lab import use_task, Task


    @solara.component
    def Page():
        number = solara.use_reactive(4)

        def square():
            time.sleep(1)
            return number.value**2

        result: Task[int] = use_task(square, dependencies=[number.value])

        solara.InputInt("Square", value=number, continuous_update=True)
        if result.finished:
            solara.Success(f"Square of {number} == {result.value}")
        solara.ProgressLinear(result.pending)
    ```

    ### Running in an asyncio task

    Note that the only difference is our function is now a coroutine function,
    and we use `asyncio.sleep` instead of `time.sleep`.

    ```solara
    import asyncio
    import solara
    from solara.lab import use_task, Task


    @solara.component
    def Page():
        number = solara.use_reactive(4)

        async def square():
            await asyncio.sleep(1)
            return number.value**2

        result: Task[int] = use_task(square, dependencies=[number.value])

        solara.InputInt("Square", value=number, continuous_update=True)
        if result.finished:
            solara.Success(f"Square of {number} == {result.value}")
        solara.ProgressLinear(result.pending)
    ```

    ## Arguments

    - `f`: The function or coroutine to run as a task.
    - `dependencies`: A list of dependencies that will trigger a rerun of the task when changed.
    - `raise_error`: If true, an error in the task will be raised. If false, the error should be handled by the
        user and is available in the `.exception` attribute of the task result object.
    - `prefer_threaded` - bool: Will run coroutine functions as a task in a thread when threads are available.
        This ensures that even when a coroutine functions calls a blocking function the UI is still responsive.
        On platform where threads are not supported (like Pyodide / WASM / Emscripten / PyScript), a coroutine
        function will always run in the current event loop.


    """

    def wrapper(f):
        task_instance = solara.use_memo(lambda: task(f, prefer_threaded=prefer_threaded), dependencies=[])
        # we always update the function so we do not have stale data in the function
        task_instance.function = f  # type: ignore

        def _prestart():
            if dependencies is not None:
                # we do not want to be in a state of .finished when the dependencies change
                # otherwise user code might render a stale value with the new dependencies
                task_instance._prestart()

        solara.use_memo(_prestart, dependencies=dependencies)

        def run():
            if dependencies is not None:
                # but we only want to execute it as an effect, which makes
                # sure that if the user assigns to a task object, the function f
                # starts after the assignment is executed
                task_instance()

        solara.use_effect(run, dependencies=dependencies)
        if raise_error:
            if task_instance.error:
                raise task_instance.exception
        return task_instance

    if f is None:
        return wrapper
    else:
        return wrapper(f)



"""# component_vue

"""
import solara
import solara.autorouting
from solara.website.utils import apidoc

from . import NoPage

title = "component_vue"
Page = NoPage
__doc__ += apidoc(solara.component_vue)  # type: ignore



# flake8: noqa
import solara
import solara as sol
from solara.alias import rv, rw

from .docutils import IncludeComponent

file_path = "yolov3.weights"
url = "https://pjreddie.com/media/files/yolov3.weights"
expected_size = 248007048


@solara.component
def DownloadFile(file_path=file_path, url=url, expected_size=expected_size, on_done=None):
    download = solara.hooks.use_download(file_path, url, expected_size=expected_size)
    downloaded_size = download.progress * expected_size
    if on_done:
        on_done(download.progress == 1)
    if download.value:
        status = "Done 🎉"
    else:
        MEGABYTES = 2.0**20.0
        status = "Downloading {}... ({:6.2f}/{:6.2f} MB)".format(file_path, downloaded_size / MEGABYTES, expected_size / MEGABYTES)
    # status = "hi"
    # return MarkdownIt(f'{status}')
    assert download.progress is not None
    with rv.Container() as main:
        # with w.VBox() as main:
        with rv.Row():
            with rv.Col(cols=1):
                progressbar = rv.ProgressLinear(value=download.progress * 100, color="primary", striped=True, height=20)
            # with rv.Col(cols=1):
            #     MarkdownIt(f'{status}')
    return main


@solara.component
def DocUseDownload():
    with rv.Container() as main:
        with rw.VBox(layout={"padding": "20px", "max_width": "1024px"}):
            solara.MarkdownIt(
                """
# use_download

```python
def use_state(initial: T, key: str = None) -> Tuple[T, Callable[[T], T]]:
    ...
```

use_state can be used to create a variable that is local to this component, and will be preserved during rerenders.

It returns a tuple with the current value, and a setter function that should be called to change the variable. A call to this setter
will trigger a rerender, and will cause the `use_state` function to return the new value on the next render.

## Simple examples

### Click button

Lets start with a Button, that renders how many times it is clicked.
        """
            )
            IncludeComponent(
                DownloadFile,
                """
import reacton
import solara.ipywidgets as w

""",
                highlight=[6],
            )
    #             MarkdownIt("""
    # ### Markdown editor
    # Lets continue with a more typical pattern, and create new new markdown component
    #         """)
    #             IncludeComponent(MarkdownIt, md_text="# This is a custom\nMark-*down* **component**")

    #             MarkdownIt("""This component does not have state itself, the markdown text can only be set via its argument.
    # A common pattern is then to have its parent component manage the state, and pass it down:
    # """)
    #             # IncludeComponent(MarkdownEditor, md="# Edit me\nand the markdown component **will** *update*", highlight=[3,5,6])
    #             MarkdownIt("""Here we see the `MarkdownEditor` component using the `use_state` function to store the markdown text, while letting the `Textarea` component change its value""")
    return main


DownloadFile("yolov3.weights", "https://pjreddie.com/media/files/yolov3.weights", expected_size=248007048)



from typing import TypeVar

from solara.toestand import Reactive

__all__ = ["reactive", "Reactive"]

T = TypeVar("T")


def reactive(value: T) -> Reactive[T]:
    """Creates a new Reactive object with the given initial value.

    Reactive objects are mostly used to manage global or application-wide state in
    Solara web applications. They provide an easy-to-use mechanism for keeping
    track of the changing state of data and for propagating those changes to
    the appropriate UI components. For managing local or component-specific
    state, consider using the [`solara.use_state()`](/api/use_state) function.


    Reactive variables can be accessed using the `.value` attribute. To modify
    the value, you can either set the `.value` property directly or use the
    `.set()` method. While both approaches are equivalent, the `.set()` method
    is particularly useful when you need to pass it as a callback function to
    other components, such as a slider's `on_value` callback.

    When a component uses a reactive variable, it
    automatically listens for changes to the variable's value. If the value
    changes, the component will automatically re-render to reflect the updated
    state, without the need to explicitly subscribe to the variable.

    Reactive objects in Solara are also context-aware, meaning that they can
    maintain separate values for each browser tab or user session. This enables
    each user to have their own independent state, allowing them to interact
    with the web application without affecting the state of other users.

    Args:
        value (T): The initial value of the reactive variable.

    Returns:
        Reactive[T]: A new Reactive object with the specified initial value.

    Example:

    ```python
    >>> counter = solara.reactive(0)
    >>> counter.value
    0
    >>> counter.set(1)
    >>> counter.value
    1
    >>> counter.value += 1
    >>> counter.value
    2
    ```


    ## Solara example

    Here's an example that demonstrates the use of reactive variables in Solara components:

    ```solara
    import solara

    counter = solara.reactive(0)

    def increment():
        counter.value += 1


    @solara.component
    def CounterDisplay():
        solara.Info(f"Counter: {counter.value}")


    @solara.component
    def IncrementButton():

        solara.Button("Increment", on_click=increment)


    @solara.component
    def Page():
        IncrementButton()
        CounterDisplay()
    ```

    In this example, we create a reactive variable counter with an initial value of 0.
    We define two components: `CounterDisplay` and `IncrementButton`. `CounterDisplay` renders the current value of counter,
    while `IncrementButton` increments the value of counter when clicked.
    Whenever the counter value changes, `CounterDisplay` automatically updates to display the new value.

    """
    return Reactive(value)



import ipywidgets as widgets

clicks = 0


def on_click(button):
    global clicks
    clicks += 1
    button.description = f"Clicked {clicks} times"


button = widgets.Button(description="Clicked 0 times")
button.on_click(on_click)



"""This Page takes an extra argument, meaning that it can cache urls like /tabular/titanic
and pass the last part of the url as argument to the Page component, so we can render content
dynamically.
"""

from typing import Optional

import solara

from .. import data
from ..components import data as data_components


@solara.component
def Page(name: Optional[str] = None, page: int = 0, page_size=100):
    if name is None or name not in data.dfs:
        with solara.Column() as main:
            solara.Title("Solara demo » table view")
            data_components.Overview()
        return main

    df = data.dfs[name].df
    with solara.ColumnsResponsive(12) as main:
        with solara.Link("/tabular"):
            solara.Text("« Back to overview")
        solara.DataTable(df=df, page=page, items_per_page=page_size)
        with solara.Head():
            solara.Title(f"Solara demo » table view » {name}")
    return main



import warnings
from typing import Any, Callable, Dict, List, Union

import reacton
import reacton.ipyvuetify as v

import solara
import solara.widgets
from solara.util import _combine_classes

Navigator = reacton.core.ComponentWidget(solara.widgets.Navigator)
GridDraggable = reacton.core.ComponentWidget(solara.widgets.GridLayout)
# keep the old name for a while
GridLayout = GridDraggable


@solara.component
def ListItem(title, icon_name: str = None, children=[], value=None):
    if value is None:
        value = title
    if children:
        with v.ListItemContent() as main:
            v.ListItemTitle(children=[title])
            if icon_name is not None:
                with v.ListItemIcon():
                    v.Icon(children=[icon_name or ""])
        return v.ListGroup(children=children, v_slots=[{"name": "activator", "children": main}], no_action=True, value=True, append_icon=icon_name)
    else:
        with v.ListItem(value=value) as main:
            if icon_name is not None:
                with v.ListItemIcon():
                    v.Icon(children=[icon_name or ""])
            with v.ListItemContent():
                v.ListItemTitle(children=[title])
        return main


def ui_dropdown(label, value=None, options=["foo", "bar"], key=None, disabled=False, **kwargs):
    key = key or str(value) + str(label) + str(options)
    value, set_value = solara.use_state(value, key)

    def set_index(index):
        set_value(options[index])

    v.Select(v_model=value, label=label, items=options, on_v_model=set_value, clearable=True, disabled=disabled, **kwargs)
    return value


def ui_text(label, value="", key=None, clearable=False, hint="", disabled=False, **kwargs):
    key = key or str(value) + str(label) + str(hint)
    value, set_value = solara.use_state(value, key)
    v.TextField(v_model=value, label=label, on_v_model=set_value, clearable=clearable, hint=hint, disabled=disabled, **kwargs)
    return value


def ui_checkbox(label, value=True, key=None, disabled=False, **kwargs):
    key = key or str(value) + str(label)
    value, set_value = solara.use_state(value, key)
    v.Checkbox(v_model=value, label=label, on_v_model=set_value, **kwargs)
    return value


def ui_slider(value=1, label="", min=0, max=100, key=None, tick_labels=None, thumb_label=None, disabled=False, **kwargs):
    key = key or str(value) + str(label)
    value, set_value = solara.use_state(value, key)
    v.Slider(
        v_model=value,
        label=label,
        min=min,
        max=max,
        on_v_model=set_value,
        ticks=tick_labels is not None,
        tick_labels=tick_labels,
        thumb_label=thumb_label,
        disabled=disabled,
        **kwargs,
    )
    return value


@solara.component
def Text(text, style: Union[str, Dict[str, str], None] = None, classes: List[str] = []):
    style_flat = solara.util._flatten_style(style)
    return v.Html(tag="span", class_=_combine_classes(classes), style_=style_flat, children=[text])


@solara.component
def Div(children=[], classes: List[str] = [], style: Union[str, Dict[str, str], None] = None, **kwargs):
    style_flat = solara.util._flatten_style(style)
    classes = classes.copy()
    kwargs = kwargs.copy()
    if "class_" in kwargs:
        classes.append(kwargs.pop("class_"))
    if "style_" in kwargs:
        style_flat += kwargs.pop("style_")
    class_ = _combine_classes(classes)

    return v.Html(tag="div", children=children, class_=class_, style_=style_flat, **kwargs)


@solara.component
def Preformatted(text, **kwargs):
    return v.Html(tag="pre", children=[text], **kwargs)


@solara.component
def IconButton(icon_name: str = None, on_click=Callable[[], None], children: list = [], click_event="click", **kwargs):
    return solara.Button(icon_name=icon_name, on_click=on_click, children=children, icon=True, click_event=click_event, **kwargs)


@solara.component
def HTML(tag="div", unsafe_innerHTML=None, style: str = None, classes: List[str] = [], attributes=None, class_: str = None):
    """Render an HTML tag with optional raw HTML text inside.

    # Arguments

     * `tag`: HTML tag name for the top level element (default: `div`)
     * `unsafe_innerHTML`: HTML string to be rendered inside the tag.
        Note that this is not sanitized, so be careful this cannot include JavaScript from user input!
     * `style`: CSS style string to be applied to the top level element.
     * `classes`: List of CSS classes to be applied to the top level element.
     * `attributes`: Dictionary of attributes to be applied to the top level element.
     * `class_`: (deprecated) CSS class to be applied to the top level element.

    """
    if attributes is None:
        attributes = {}
    else:
        attributes = attributes.copy()
    if style:
        attributes["style"] = style
    if class_ or classes:
        class_ = _combine_classes([*classes, *([] if class_ is None else [class_])])
        attributes["class"] = class_
    return solara.widgets.HTML.element(tag=tag, unsafe_innerHTML=unsafe_innerHTML, attributes=attributes)


@solara.component
def VBox(children=[], grow=True, align_items="stretch", classes: List[str] = []):
    """Deprecated. Use `Row` instead."""
    style = f"flex-direction: column; align-items: {align_items};"
    if grow:
        style += "flex-grow: 1;"
    class_ = _combine_classes(["d-flex", *classes])
    return v.Sheet(class_=class_, style_=style, elevation=0, children=children)


@solara.component
def HBox(children=[], grow=True, align_items="stretch", classes: List[str] = []):
    """Deprecated. Use `Column` instead."""
    style = f"flex-direction: row; align-items: {align_items}; "
    if grow:
        style += "flex-grow: 1;"
    class_ = _combine_classes(["d-flex", *classes])
    return v.Sheet(class_=class_, style_=style, elevation=0, children=children)


@solara.component
def Row(children=[], gap="12px", justify="start", margin: int = 0, classes: List[str] = [], style: Union[str, Dict[str, str], None] = None):
    """Lays out children in a row, side by side, with the given gap between them.

    See also [Column](/api/column).

    Example with three children side by side:

    ```solara
    import solara

    @solara.component
    def Page():
        with solara.Row(gap="10px", justify="space-around"):
            solara.Text("On the left")
            solara.Text("In the middle")
            solara.Text("On the right")
    ```
    ## Arguments

     * `children`: List of children to render in the column.
     * `gap`: The gap between each child, as a CSS string.
     * `justify`: How children are distributed along the x/horizontal-axis, can be "start" (default), "center", "end", "space-around",
        "space-between" or "space-evenly".
        (*Note: this translates to justify-content in CSS*).
     * `margin`: The margin around the column, translate to 4*margin pixels.
     * `classes`: List of CSS classes to apply to the column.
     * `style`: CSS style to apply to the column.

    """
    align_items = "stretch"
    style_flat = solara.util._flatten_style(style)
    style_flat = f"flex-direction: row; align-items: {align_items}; justify-content: {justify}; column-gap: {gap};" + style_flat + ";"
    # valid css values, but we don't list them as options to avoid confusion
    extra_justify_options = ["left", "right", "flex-start", "flex-end"]
    if justify not in (["start", "center", "end", "space-around", "space-between", "space-evenly"] + extra_justify_options):
        warnings.warn(f"Invalid value for justify: {justify}, possible values are: start, center, end, space-around, space-between, space-evenly")
    class_ = _combine_classes(["d-flex", f"ma-{margin}", *classes])
    return v.Sheet(class_=class_, style_=style_flat, elevation=0, children=children)


@solara.component
def Column(children=[], gap="12px", align="stretch", margin: int = 0, classes: List[str] = [], style: Union[str, Dict[str, str], None] = None):
    """Lays out children in a column on top of each other, with the given gap between them.

    See also [Row](/api/row).

    Example with three children on top of each other:

    ```solara
    import solara

    @solara.component
    def Page():
        with solara.Column(gap="10px"):
            solara.Text("On top")
            solara.Text("In the middle")
            solara.Text("On bottom")
    ```

    ## Arguments

     * `children`: List of children to render in the column.
     * `gap`: The gap between each child, as a CSS string.
     * `align`: The alignment of the children, can be "start", "center", "end", "stretch" (default).
        (*Note: this translates to align-items in CSS*).
     * `margin`: The margin around the column, translate to 4*margin pixels.
     * `classes`: List of CSS classes to apply to the column.
     * `style`: CSS style to apply to the column.

    """
    if align == "left":
        warnings.warn("align='left' does not exists, you probably want align='start' instead")
        align = "start"
    if align == "right":
        warnings.warn("align='right' does not exists, you probably want align='end' instead")
        align = "end"
    # valid css options, but we don't list them as options to avoid confusion
    extra_align_options = ["flex-start", "flex-end", "self-start", "self-end", "baseline"]
    if align not in (["start", "center", "end", "stretch"] + extra_align_options):
        warnings.warn(f"Invalid value for align: {align}, possible values are 'start', 'center', 'end' or 'stretch'")
    style_flat = solara.util._flatten_style(style)
    style_flat = f"flex-direction: column; align-items: {align}; row-gap: {gap};" + style_flat + ";"
    class_ = _combine_classes(["d-flex", f"ma-{margin}", *classes])
    return v.Sheet(class_=class_, style_=style_flat, elevation=0, children=children)


@solara.component
def GridFixed(columns=4, column_gap="10px", row_gap="10px", children=[], align_items="stretch", justify_items="stretch"):
    """

    See css grid spec:
    https://css-tricks.com/snippets/css/complete-guide-grid/
    """
    style = (
        f"display: grid; grid-template-columns: repeat({columns}, minmax(0, 1fr)); "
        + f"grid-column-gap: {column_gap}; grid-row-gap: {row_gap}; align-items: {align_items}; justify-items: {justify_items}"
    )
    return Div(style_=style, children=children)


@solara.component
def Padding(size, children=[], grow=True):
    style = "flex-direction: row;"
    if grow:
        style += "flex-grow: 1;"
    return v.Sheet(class_=f"pa-{size}", style_=style, elevation=0, children=children)


@solara.component
def FigurePlotly(
    fig,
    on_selection: Callable[[Any], None] = None,
    on_deselect: Callable[[Any], None] = None,
    on_click: Callable[[Any], None] = None,
    on_hover: Callable[[Any], None] = None,
    on_unhover: Callable[[Any], None] = None,
    on_relayout: Callable[[Any], None] = None,
    dependencies=None,
):
    from plotly.graph_objs._figurewidget import FigureWidget

    def on_points_callback(data):
        if not data:
            return

        event_type = data["event_type"]
        event_mapping = {
            "plotly_click": on_click,
            "plotly_hover": on_hover,
            "plotly_unhover": on_unhover,
            "plotly_selected": on_selection,
            "plotly_deselect": on_deselect
        }
        
        callback = event_mapping.get(event_type)
        if callback:
            callback(data)

    fig_element = FigureWidget.element(
        on__js2py_pointsCallback=on_points_callback,
        on__js2py_relayout=on_relayout
    )

    def update_data():
        fig_widget: FigureWidget = solara.get_widget(fig_element)
        fig_widget.layout = fig.layout

        length = len(fig_widget.data)
        fig_widget.add_traces(fig.data)
        data = list(fig_widget.data)
        fig_widget.data = data[length:]

    solara.use_effect(update_data, dependencies or fig)
    return fig_element


@solara.component
def Code(path, path_header=None):
    path_header = path_header or path
    with open(path) as f:
        code = f.read()
    md = solara.Markdown(
        f"""
### {path_header}

```python
{code}
```

"""
    )

    with v.ExpansionPanels() as main:
        with v.ExpansionPanel():
            with v.ExpansionPanelHeader(children=["View source"]):
                pass
            with v.ExpansionPanelContent(children=[md]):
                pass
    return main



"""Demonstrates how to use Vue components in Solara.

Although many components can be made from the Python side, sometimes it is easier to write components using Vue directly.
It can also be beneficial for performance, since instead of creating many widgets from the Python side we only send data to
the frontend. If event handling is also done on the frontend, this reduces latency and makes you app feel much smoother.

See [the API documentation on component_vue](/api/component_vue) for more information.

This example is based on [the vuetify docs](https://v2.vuetifyjs.com/en/components/sparklines/#custom-labels),
Note that the "Go to report" button does not do anything yet.

"""

from typing import Callable

import numpy as np

import solara

seed = solara.reactive(42)


@solara.component_vue("mycard.vue")
def MyCard(
    event_goto_report: Callable[[dict], None],
    value=[1, 10, 30, 20, 3],
    caption="My Card",
    color="red",
):
    pass


@solara.component
def Page():
    gen = np.random.RandomState(seed=seed.value)
    sales_data = np.floor(np.cumsum(gen.random(7) - 0.5) * 100 + 100)
    show_report = solara.use_reactive(False)

    with solara.Column(style={"min-width": "600px"}):
        if show_report.value:
            with solara.Card("Report"):
                solara.Markdown("Lorum ipsum dolor sit amet")
                solara.Button("Go back", on_click=lambda: show_report.set(False))
        else:

            def new_seed():
                seed.value = np.random.randint(0, 100)

            solara.Button("Generate new data", on_click=new_seed)

            MyCard(value=sales_data.tolist(), color="green", caption="Sales Last 7 Days", event_goto_report=lambda data: show_report.set(True))



"""
# ToggleButtons

ToggleButtons are in two flavours, for single, and for multiple selections.


"""
import solara
from solara.website.utils import apidoc

from . import NoPage

title = "ToggleButtons"

Page = NoPage


__doc__ += "# ToggleButtonsSingle"
__doc__ += apidoc(solara.ToggleButtonsSingle.f)  # type: ignore
__doc__ += "# ToggleButtonsMultiple"
__doc__ += apidoc(solara.ToggleButtonsMultiple.f)  # type: ignore



"""
# Sidebar

"""
import solara
import solara.lab
from solara.website.utils import apidoc


@solara.component
def Page():
    return solara.Markdown(
        """

    The sidebar can only be shown in embedded mode on this page.
    Visit the [Scatter app demo](/apps/scatter) to see an example of a full sidebar used in Soalra server.

    [![AppLayout screenshot](https://dxhl76zpt6fap.cloudfront.net/public/docs/app-layout.webp)](/apps/scatter)
    """
    )


__doc__ += apidoc(solara.Sidebar.f)  # type: ignore



"""# Head
"""

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    with solara.VBox() as main:
        solara.Info("A Head component does not render somesome visual on the page, but it is used to avoid duplicate tags, such as titles.")
        with solara.Head():
            # title should always occur inside a Head component
            solara.Title("Custom title")

    return main


__doc__ += apidoc(solara.Head.f)  # type: ignore



from typing import Callable, Dict, Optional, Tuple, cast

import ipyvuetify as vy
import reacton.core
import traitlets

import solara
import solara.lab


class TitleWidget(vy.VuetifyTemplate):
    template_file = (__file__, "title.vue")
    title = traitlets.Unicode().tag(sync=True)
    level = traitlets.Int().tag(sync=True)


Titles = Dict[str, Tuple[int, str]]


def _set_titles_default(updater: Callable[[Titles], Titles]):
    pass


titles_context = solara.create_context(_set_titles_default)


def use_title_get() -> Optional[str]:
    titles, set_titles = solara.use_state(cast(Titles, {}))
    titles_context.provide(set_titles)  # type: ignore
    if titles:
        title = max([(order, title) for (key, (order, title)) in titles.items()], key=lambda x: x[0])[1]
    else:
        title = None
    return title


def use_title_set(title: str, offset: int):
    key = solara.use_unique_key(prefix="title-")
    set_titles = solara.use_context(titles_context)

    def update():
        set_titles(lambda titles: {**titles, key: (offset, title)})

    solara.use_effect(update, [title])

    def restore():
        def cleanup():
            def without(titles):
                titles_restored = titles.copy()
                titles_restored.pop(key, None)
                return titles_restored

            set_titles(without)

        return cleanup

    solara.use_effect(restore, [])


@solara.component
def Title(title: str):
    """Set the title of a page.

    ```python
    import solara

    @solara.component
    def Page():
        with solara.VBox() as main:
            MyAwesomeComponent()
            solara.Title("My page title")
        return main
    ```

    If multiple Title components are used, the 'deepest' child will take precedence.

    ## Arguments

     * title: the title of the page
    """
    level = 0
    rc = reacton.core.get_render_context()
    context = rc.context
    while context and context.parent:
        level += 1
        context = context.parent
    offset = 2**level
    use_title_set(title, offset)

    return TitleWidget.element(title=title, level=level)



"""
# SqlCode
"""
import os
import sqlite3
import threading
from typing import Optional, cast

import pandas as pd

try:
    import vaex.datasets
except ImportError:
    vaex = None

import solara
from solara.alias import rv
from solara.website.utils import apidoc

if vaex is not None:
    df_iris = vaex.datasets.iris().to_pandas_df()
    df_titanic = vaex.datasets.titanic().to_pandas_df()

    def create_db():
        conn = sqlite3.connect(filename)
        df_iris.to_sql("iris", conn, if_exists="replace", index=False)
        df_titanic.to_sql("titanic", conn, if_exists="replace", index=False)
        conn.close()

    filename = "solara-sql.db"
    if not os.path.exists(filename):
        create_db()

    conn = sqlite3.connect(filename)
    conn.row_factory = sqlite3.Row
    table_names = [k[0] for k in conn.execute("SELECT name FROM sqlite_master WHERE type='table' ORDER BY name").fetchall()]
    table_hints = {table_name: conn.execute(f"SELECT * from {table_name} LIMIT 1").fetchone().keys() for table_name in table_names}


@solara.component
def Page():
    if vaex is None:
        return solara.Error("Vaex is not installed, run pip install vaex-core vaex-hdf5")
    query, set_query = solara.use_state("SELECT * from titanic")
    query_executed, set_query_executed = solara.use_state(cast(Optional[str], None))

    def run_query(cancel: threading.Event) -> pd.DataFrame:
        if not query_executed:
            return
        conn = sqlite3.connect(filename)
        cursor = conn.cursor()
        cursor.execute(query_executed)

        df = pd.DataFrame(cursor.fetchall(), columns=[k[0] for k in cursor.description])
        return df

    result: solara.Result[pd.DataFrame] = solara.use_thread(run_query, dependencies=[query_executed])
    with solara.VBox() as main:
        solara.SqlCode(query=query, tables=table_hints, on_query=set_query)
        enable_execute = (query != query_executed) or result.error is not None

        def execute():
            set_query_executed(query)
            if query == query_executed and result.error:
                result.retry()  # type: ignore

        solara.Button("Execute", on_click=execute, disabled=not enable_execute)
        if result.error:
            solara.Error(f"Ooops {result.error}")
        elif not query:
            solara.Info("No query")

        elif result.value is not None:
            solara.Markdown(f"Result for query: `{query_executed}`")
            df = result.value
            solara.DataFrame(df)
        elif query_executed is not None:
            with solara.Div():
                solara.Text("Loading data...")
                rv.ProgressCircular(indeterminate=True, class_="solara-progress")
    return main


__doc__ += apidoc(solara.SqlCode.f)  # type: ignore



import solara


@solara.component
def Page():
    with solara.Link("/showcase"):
        solara.Text("« Back to Showcases")
    with solara.ColumnsResponsive(12, medium=6):
        with solara.Column(align="end", style={"height": "100%", "justify-content": "center"}):
            solara.Markdown(
                """
                Team 2 built a travel assistant, which utilized various third party APIs to display rich information about the desired destination
                during the dates selected for a visit. The assistant also provided a list of recommended activities, and a map of the area.

                Take a look at the project on [Ploomber](https://jolly-moon-5966.ploomberapp.io/)
                """
            )
        with solara.v.Html(tag="video", attributes={"controls": "controls", "autoplay": "autoplay"}, style_="width:100%;"):
            solara.v.Html(
                tag="source",
                attributes={"src": "https://dxhl76zpt6fap.cloudfront.net/public/showcase/solarathon_2023_team_2/preview.webm", "type": "video/webm"},
            )



"""
# Chat Components

These are the components available in Solara to build chat interfaces.
Although they can be used together to create a pre-built chat interface, inserting custom components is also possible.
For an example of how to use a custom message component, see the bottom of the page.

Also check out the [Chatbot](/examples/ai/chatbot) example.

# ChatBox
"""
import solara
from solara.website.utils import apidoc

from . import NoPage

title = "Chat Components"


__doc__ += apidoc(solara.lab.components.chat.ChatBox.f)  # type: ignore
__doc__ += "# ChatInput"
__doc__ += apidoc(solara.lab.components.chat.ChatInput.f)  # type: ignore
__doc__ += "# ChatMessage"
__doc__ += apidoc(solara.lab.components.chat.ChatMessage.f)  # type: ignore
__doc__ += """## Different Message Styles

The `ChatMessage` component has a few different styles available. These are shown below.

```solara
import solara

@solara.component
def Page():
    with solara.Column():
        solara.lab.ChatMessage(["Default"])
        solara.lab.ChatMessage(["`color`"], color="#ff991f")
        solara.lab.ChatMessage(["`avatar_background_color`"], avatar_background_color="success")
        solara.lab.ChatMessage(["`border_radius='20px'`"], border_radius="20px")
        solara.lab.ChatMessage(["`notch=True`"], notch=True)
```
"""

__doc__ += """
# A Basic Example


```solara
import solara

messages = solara.reactive([])
name = solara.reactive("User")

def send(new_message):
    messages.set([
        *messages.value,
        {"user": True, "name": name.value, "message": new_message,},
    ])

@solara.component
def Page():
    solara.InputText("username", value=name)
    with solara.Column(style={"min-height": "50vh"}):
        with solara.lab.ChatBox():
            for item in messages.value:
                with solara.lab.ChatMessage(
                    user=item["user"],
                    name=item["name"],
                ):
                    solara.Markdown(item["message"])
        solara.lab.ChatInput(send_callback=send)
```

# A Custom Message Component

```solara
import solara

messages = solara.reactive([])

def send(new_message):
    messages.set([
        *messages.value,
        {"name": "User", "message": new_message},
    ])

@solara.component
def CustomMessage(
    message,
    name = "User",
):
    with solara.Column(gap=0):
        with solara.Row():
            solara.Markdown(f"**{name}**: {message}")
        solara.v.Divider(style_="margin: 0;")

@solara.component
def Page():
    with solara.Column(style={"min-height": "50vh"}):
        with solara.lab.ChatBox():
            for item in messages.value:
                CustomMessage(
                    message=item["message"],
                    name=item["name"],
                )
        solara.lab.ChatInput(send_callback=send)
```
"""

Page = NoPage



"""Renders a plotly figure object and support for events.

You can use a regular plotly figure, for instance using plotly.express.

Supported events are include in the example below.

For performance, a dependency list can be provided. Any change to the dependency list will cause the Figure data to be updated.
This can be significantly faster than comparing he figure data directly.

"""

import plotly.express as px
import solara

df = px.data.iris()


@solara.component
def Page():
    with solara.VBox() as main:
        selection_data, set_selection_data = solara.use_state(None)
        click_data, set_click_data = solara.use_state(None)
        hover_data, set_hover_data = solara.use_state(None)
        unhover_data, set_unhover_data = solara.use_state(None)
        deselect_data, set_deselect_data = solara.use_state(None)
        fig = px.scatter(df, x="sepal_width", y="sepal_length", color="species")
        solara.FigurePlotly(
            fig, on_selection=set_selection_data, on_click=set_click_data, on_hover=set_hover_data, on_unhover=set_unhover_data, on_deselect=set_deselect_data
        )

        solara.Markdown(
            f"""
# Events data
## selection
```
{selection_data}
```

## click
```
{click_data}
```

## hover
```
{hover_data}
```

## unhover
```
{unhover_data}
```

## deselect
```
{deselect_data}
```


"""
        )
    return main



import dataclasses
import os

import ipyvue
import traitlets

import solara
from solara.alias import rw


@dataclasses.dataclass
class Clicks:
    value: int


@solara.component
def ButtonClick():
    clicks, set_clicks = solara.use_state(Clicks(0))
    return rw.Button(description=f"Clicked {clicks.value} times", on_click=lambda: set_clicks(Clicks(clicks.value + 1)))


app = ButtonClick()


@solara.component
def ClickBoom():
    count, set_count = solara.use_state(0)
    if count == 1:
        raise ValueError("I crash on 1")
    return solara.Button("Boom", on_click=lambda: set_count(count + 1))


clickboom = ClickBoom()


class TestWidget(ipyvue.VueTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "test.vue"))

    value = traitlets.Any(0).tag(sync=True)


@solara.component
def VueTestApp():
    return TestWidget.element(value="foobar")


vue_test_app = VueTestApp()



"""# CrossFilterSelect
"""
import plotly

import solara
import solara.lab
from solara.website.utils import apidoc

df = plotly.data.tips()


@solara.component
def Page():
    solara.provide_cross_filter()
    with solara.VBox() as main:
        solara.CrossFilterReport(df, classes=["py-2"])
        solara.CrossFilterSelect(df, "sex")
        solara.CrossFilterSelect(df, "time")
    return main


__doc__ += apidoc(solara.CrossFilterSelect.f)  # type: ignore



import functools
import typing
from typing import Callable

import numpy as np
import plotly.express as px
import plotly.graph_objs as go
import typing_extensions

# behave as the px module
from plotly.express import *  # noqa: F401, F403
from plotly.express._core import make_figure

import solara

P = typing_extensions.ParamSpec("P")
T = typing.TypeVar("T")
T2 = typing.TypeVar("T2")


# patch histogram until issue 3859 in plotly is resolved


def histogram(
    data_frame=None,
    x=None,
    y=None,
    custom_data=None,
    color=None,
    pattern_shape=None,
    facet_row=None,
    facet_col=None,
    facet_col_wrap=0,
    facet_row_spacing=None,
    facet_col_spacing=None,
    hover_name=None,
    hover_data=None,
    animation_frame=None,
    animation_group=None,
    category_orders=None,
    labels=None,
    color_discrete_sequence=None,
    color_discrete_map=None,
    pattern_shape_sequence=None,
    pattern_shape_map=None,
    marginal=None,
    opacity=None,
    orientation=None,
    barmode="relative",
    barnorm=None,
    histnorm=None,
    log_x=False,
    log_y=False,
    range_x=None,
    range_y=None,
    histfunc=None,
    cumulative=None,
    nbins=None,
    text_auto=False,
    title=None,
    template=None,
    width=None,
    height=None,
) -> go.Figure:
    """
    In a histogram, rows of `data_frame` are grouped together into a
    rectangular mark to visualize the 1D distribution of an aggregate
    function `histfunc` (e.g. the count or sum) of the value `y` (or `x` if
    `orientation` is `'h'`).
    """
    return make_figure(
        args=locals(),
        constructor=go.Histogram,
        trace_patch=dict(
            histnorm=histnorm,
            histfunc=histfunc,
            cumulative=dict(enabled=cumulative),
        ),
        layout_patch=dict(barmode=barmode, barnorm=barnorm),
    )


px.histogram = histogram


def _make_self_describing(f: Callable[P, T]):
    # put the arguments on the function and the function into the return value
    if hasattr(f, "__wrapped__"):
        # on hot reload, we need to unwrap the function to not wrap it twice
        f = f.__wrapped__

    @functools.wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs):
        fig = f(*args, **kwargs)
        fig._func = f  # type: ignore
        fig._args = args  # type: ignore
        fig._kwargs = kwargs  # type: ignore
        return fig

    return wrapper


def _patch():
    for name in dir(px):
        f = getattr(px, name)
        if callable(f):
            f = _make_self_describing(f)
            setattr(px, name, f)


_patch()


@solara.component
def CrossFilteredFigurePlotly(fig):
    first_arg_name = fig._func.__code__.co_varnames[0]
    kwargs = fig._kwargs.copy()
    kwargs.update(zip(fig._func.__code__.co_varnames, fig._args))

    df = kwargs[first_arg_name]
    name = fig._func.__name__
    filter, set_filter = solara.use_cross_filter(id(df), name)
    dff = df
    if filter is not None:
        dff = df[filter]
        for key, value in kwargs.items():
            if not isinstance(value, str):
                try:
                    n = len(value)
                except TypeError:
                    pass  # int or bool, or anything without a length
                except Exception:
                    raise
                else:
                    if n == len(df):
                        kwargs[key] = np.array(value)[filter]
        original_indices = np.arange(len(df), dtype="int64")
    else:
        original_indices = None

    kwargs[first_arg_name] = dff

    index = np.arange(len(dff))
    custom_data = [index]
    has_custom_data = "custom_data" in kwargs

    if has_custom_data:
        kwargs["custom_data"] = custom_data + list(kwargs["custom_data"])
    else:
        kwargs["custom_data"] = custom_data
    new_fig = fig._func(**kwargs)
    new_fig.layout = fig.layout

    # the df is split into multiple traces, generate the data such that we can transform back from trace+point indices to df.index
    indices = []
    offset = 0
    index_offsets_list = []
    for data in new_fig.data:
        indices.append(data.customdata.T[0])
        index_offsets_list.append(offset)
        offset += len(data.customdata.T[0])
    index_offsets = np.array(index_offsets_list)
    # strip of the custom data we added
    for data in new_fig.data:
        if has_custom_data:
            data.customdata = data.customdata.T[len(custom_data) :].T
        else:
            data.customdata = None

    def on_selection(data):
        if data is not None:
            trace_indexes = np.array(data["points"]["trace_indexes"])
            point_indexes = np.array(data["points"]["point_indexes"])
            if len(trace_indexes):
                indices_selected = index_offsets[trace_indexes] + point_indexes

                if filter is not None:
                    assert original_indices is not None
                    # these are references to the filtered dataframe
                    indices_selected = original_indices[filter][indices_selected]
                mask = np.zeros(len(df), dtype=bool)
                mask[indices_selected] = True
                set_filter(mask)
            else:
                set_filter(None)

    def on_deselect(data):
        set_filter(None)

    return solara.FigurePlotly(new_fig, on_selection=on_selection, on_deselect=on_deselect)


# for backwards compatibility
FigurePlotlyCrossFiltered = CrossFilteredFigurePlotly


def _wraps(f: Callable[P, T]):
    @functools.wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs):
        fig = f(*args, **kwargs)
        return CrossFilteredFigurePlotly(fig)

    return wrapper


scatter = _wraps(px.scatter)
scatter_3d = _wraps(px.scatter_3d)
scatter_polar = _wraps(px.scatter_polar)
scatter_ternary = _wraps(px.scatter_ternary)
scatter_mapbox = _wraps(px.scatter_mapbox)
scatter_geo = _wraps(px.scatter_geo)
scatter_matrix = _wraps(px.scatter_matrix)
density_contour = _wraps(px.density_contour)
density_heatmap = _wraps(px.density_heatmap)
density_mapbox = _wraps(px.density_mapbox)
line = _wraps(px.line)
line_3d = _wraps(px.line_3d)
line_polar = _wraps(px.line_polar)
line_ternary = _wraps(px.line_ternary)
line_mapbox = _wraps(px.line_mapbox)
line_geo = _wraps(px.line_geo)
parallel_coordinates = _wraps(px.parallel_coordinates)
parallel_categories = _wraps(px.parallel_categories)
area = _wraps(px.area)
bar = _wraps(px.bar)
timeline = _wraps(px.timeline)
bar_polar = _wraps(px.bar_polar)
violin = _wraps(px.violin)
box = _wraps(px.box)
strip = _wraps(px.strip)
histogram = _wraps(px.histogram)
ecdf = _wraps(px.ecdf)
choropleth = _wraps(px.choropleth)
choropleth_mapbox = _wraps(px.choropleth_mapbox)
pie = _wraps(px.pie)
sunburst = _wraps(px.sunburst)
treemap = _wraps(px.treemap)
icicle = _wraps(px.icicle)
funnel = _wraps(px.funnel)
funnel_area = _wraps(px.funnel_area)
imshow = _wraps(px.imshow)



from pathlib import Path

import folium
import folium.plugins
import matplotlib.pyplot as plt
import pandas as pd

import solara

districts = solara.reactive(
    ["Bayview", "Northern"],
)
categories = solara.reactive(["Vandalism", "Assault", "Robbery"])
limit = solara.reactive(100)


ROOT = Path(solara.__file__).parent / "website" / "pages" / "docs" / "content" / "04-tutorial"
path = ROOT / Path("SF_crime_sample.csv.gz")
url = "https://raw.githubusercontent.com/widgetti/solara/master/solara/website/pages/docs/content/04-tutorial/SF_crime_sample.csv"

if path.exists():
    df_crime = pd.read_csv(path)
else:
    df_crime = pd.read_csv(url)

df_crime


df_crime["Category"] = df_crime["Category"].str.title()
df_crime["PdDistrict"] = df_crime["PdDistrict"].str.title()


def crime_filter(df, district_values, category_values):
    df_dist = df.loc[df["PdDistrict"].isin(district_values)]
    df_category = df_dist.loc[df_dist["Category"].isin(category_values)]
    return df_category


def crime_charts(df):
    cat_unique = df["Category"].value_counts()
    cat_unique = cat_unique.reset_index()

    dist_unique = df["PdDistrict"].value_counts()
    dist_unique = dist_unique.reset_index()

    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(20, 10))

    ax1.bar(cat_unique["Category"], cat_unique["count"])
    ax1.set_title("Amount of Criminal Case Based on Category")
    ax2.bar(dist_unique["PdDistrict"], dist_unique["count"])
    ax2.set_title("Amount of Criminal Case in Selected District")

    # this does not work on solara yet: https://github.com/widgetti/solara/issues/399
    # display(fig)
    # plt.close(fig)
    solara.FigureMatplotlib(fig)


def crime_map(df):
    latitude = 37.77
    longitude = -122.42

    sanfran_map = folium.Map(height=400, location=[latitude, longitude], zoom_start=12)

    incidents = folium.plugins.MarkerCluster().add_to(sanfran_map)

    # loop through the dataframe and add each data point to the mark cluster
    for (
        lat,
        lng,
        label,
    ) in zip(df.Y, df.X, df.Category):
        folium.Marker(
            location=[lat, lng],
            icon=None,
            popup=label,
        ).add_to(incidents)

    # show map
    solara.display(sanfran_map)


@solara.component
def View():
    dff = crime_filter(df_crime, districts.value, categories.value)
    row_count = len(dff)
    if row_count > limit.value:
        solara.Warning(f"Only showing the first {limit.value} of {row_count:,} crimes on map")
    with solara.Column(style={"max-height": "400px"}):
        crime_map(dff.iloc[: limit.value])
    if row_count > 0:
        crime_charts(dff)
    else:
        solara.Warning("You filtered out all the data, no charts shown")


@solara.component
def Controls():
    solara.SelectMultiple("District", all_values=[str(k) for k in df_crime["PdDistrict"].unique().tolist()], values=districts)  # type: ignore
    solara.SelectMultiple("Category", all_values=[str(k) for k in df_crime["Category"].unique().tolist()], values=categories)  # type: ignore
    solara.Text("Maximum number of rows to show on map")
    solara.SliderInt("", value=limit, min=1, max=1000)


@solara.component
def Page():
    with solara.Sidebar():
        Controls()
    View()


# only needed on the solara website itself
@solara.component
def Layout(children):
    route, routes = solara.use_route()
    return solara.AppLayout(children=children)



"""# GPT-4 token encoder and decoder

This example shows how to use a language model tokenizer to encode and decode and how to search for a token

Inspired by: [Understanding GPT tokenizers by Simon Willison](https://simonwillison.net/2023/Jun/8/gpt-tokenizers/)


## Note

Install pandas and tiktoken with `pip install pandas tiktoken`

"""
import random

import pandas as pd
import tiktoken

import solara
import solara.lab

# Get tokenizer for gpt-4
tokenizer = tiktoken.encoding_for_model("gpt-4")

# Create dataframe mapping token IDs and tokens
MAX_TOKENS = 50257
df = pd.DataFrame()
df["token ID"] = range(MAX_TOKENS)
df["token"] = [tokenizer.decode([i]) for i in range(MAX_TOKENS)]

sentence = solara.reactive("Example text is here")
tokens_ids_to_lookup = solara.reactive("")
tokens_filter = solara.reactive("")


@solara.component
def GithubAvatar(name: str, handle: str, img: str):
    with solara.v.Html(tag="a", attributes={"href": f"https://github.com/{handle}/", "target": "_blank"}):
        with solara.v.ListItem(class_="pa-0"):
            with solara.v.ListItemAvatar(color="grey darken-3"):
                solara.v.Img(
                    class_="elevation-6",
                    src=img,
                )
            with solara.v.ListItemContent():
                solara.v.ListItemTitle(children=["By " + name])


@solara.component
def Token(token: int):
    # create random color dependent on the position
    random.seed(token)
    random_color = "".join([random.choice("0123456789ABCDEF") for k in range(6)])
    with solara.Div(style="display: inline;"):
        with solara.Div(
            style={
                "display": "inline",
                "padding": "6px",
                "border-right": "3px solid white",
                "line-height": "3em",
                "font-family": "courier",
                "background-color": f"#{random_color}",
                "color": "white",
                "position": "relative",
            },
        ):
            solara.Text(
                " " + str(token),
                style={
                    "display": "inline",
                    "position": "absolute",
                    "top": "5.5ch",
                    "line-height": "1em",
                    "left": "-0.5px",
                    "font-size": "0.45em",
                },
            )
            solara.Text(str(tokenizer.decode([token])))


@solara.component
def Page():
    with solara.Column(margin=10):
        solara.Markdown("# GPT-4 token encoder and decoder")
        solara.Markdown("This is an educational tool for understanding how tokenization works.")
        GithubAvatar(
            "Alonso Silva Allende",
            "alonsosilvaallende",
            "https://avatars.githubusercontent.com/u/30263736?v=4",
        )

        solara.InputText("Enter text to tokenize it:", value=sentence, continuous_update=True)
        tokens = tokenizer.encode(sentence.value)
        with solara.Div(style="display: inline;"):
            for token in tokens:
                Token(token)

        solara.InputText("Or convert space separated tokens to text:", value=tokens_ids_to_lookup, continuous_update=True)
        token_input = [int(span) for span in tokens_ids_to_lookup.value.split(" ") if span != ""]
        text_output = tokenizer.decode(token_input)
        solara.Markdown(f"{text_output}")

        solara.Markdown("## Search tokens")
        solara.InputText("Search for a token:", value=tokens_filter, continuous_update=True)
        df_subset = df[df["token"].str.startswith(tokens_filter.value)]
        solara.Markdown(f"{df_subset.shape[0]:,} results")
        solara.DataFrame(df_subset, items_per_page=10)



"""
# Cookies and Headers

## Cookies

Solara provides access to cookies through `solara.lab.cookies`.

```python
cookies: Reactive[Optional[Dict[str, str]]] = reactive(cast(Optional[Dict[str, str]], None))
```

This is a reactive object that can be used to read the cookies transferred with the the initial request sent to the server by the client.
Thus it is possible to access also HTTPOnly cookies. The cookies will be updated whenever the client sends a new request to the server,
i.e. only when the client connects to a kernel.

The cookies are read into a dictionary with the cookie names as keys and the cookie values as values.

## Headers

Solara provides access to headers in the form of `solara.lab.headers`.

```python
headers: Reactive[Optional[Dict[str, List[str]]]] = reactive(cast(Optional[Dict[str, List[str]]], None))
```

`headers` concatenates the values of multiple header fields with the same name into a list of strings. However, it is still possible that multiple
values are present in one field already in the request, in which case one string will contain multiple values. Unfortunately reversing this is
highly non-trivial.

Note that the header field names are all lower-case.

```solara
import solara
from solara.lab import headers


@solara.component
def Page():
    solara.Markdown(f"Your 'user-agent' is: {headers.value['user-agent']}")

```

"""

from . import NoPage

title = "Cookies and Headers"
Page = NoPage



"""# FigureAltair

"""

import altair as alt
import pandas as pd

import solara
from solara.website.utils import apidoc

title = "FigureAltair"

df = pd.DataFrame({"a": ["A", "B", "C", "D", "E", "F", "G", "H", "I"], "b": [28, 55, 43, 91, 81, 53, 19, 87, 52]})


@solara.component
def Page():
    click_data, set_click_data = solara.use_state(None)
    hover_data, set_hover_data = solara.use_state(None)

    chart = alt.Chart(df).mark_bar().encode(x="a", y="b")

    with solara.Div() as main:
        solara.AltairChart(chart, on_click=set_click_data, on_hover=set_hover_data)

        solara.Markdown(
            f"""
Click data:

```
{click_data}
```

Hover data:
```
{hover_data}
```
            """
        )

    return main


__doc__ += apidoc(solara.FigureAltair.f)  # type: ignore
